import logging
from functools import lru_cache
from pathlib import Path

from pathspec import GitIgnoreSpec, PathSpec


logger = logging.getLogger(__name__)


@lru_cache
def get_gitignore(root: Path) -> PathSpec:
    gitignore = root / '.gitignore'

    lines = []
    if gitignore.is_file():
        with gitignore.open(encoding='utf-8') as f:
            lines = f.readlines()
    else:
        print(f'WARNING: No .gitignore found in {root}')

    return GitIgnoreSpec.from_lines(lines)



from pathlib import Path

from manageprojects.cookiecutter_generator import (
    build_dst_path,
    generate_reverse_info,
    iter_context,
    replace_path,
)
from manageprojects.tests.base import BaseTestCase


class CookiecutterGeneratorTestCase(BaseTestCase):
    def test_iter_context(self):
        result = list(
            iter_context(
                context={
                    'cookiecutter': {
                        '_template': 'https://github.com/jedie/cookiecutter_templates/',
                        'package_name': 'PyInventory',
                        'package_url': 'https://github.com/jedie/PyInventory',
                        'issues_url': 'https://github.com/jedie/PyInventory/issues',
                    }
                }
            )
        )
        self.assertEqual(
            result,
            [
                ('cookiecutter.package_name', 'PyInventory'),
                ('cookiecutter.package_url', 'https://github.com/jedie/PyInventory'),
                ('cookiecutter.issues_url', 'https://github.com/jedie/PyInventory/issues'),
            ],
        )

    def test_generate_reverse_info(self):
        reverse_info = generate_reverse_info(
            cookiecutter_context={
                'cookiecutter': {
                    '_template': 'https://github.com/jedie/cookiecutter_templates/',
                    'package_name': 'PyInventory',
                    'package_url': 'https://github.com/jedie/PyInventory',
                    'issues_url': 'https://github.com/jedie/PyInventory/issues',
                }
            }
        )
        self.assertEqual(
            reverse_info,
            (
                ('https://github.com/jedie/PyInventory/issues', '{{ cookiecutter.issues_url }}'),
                ('https://github.com/jedie/PyInventory', '{{ cookiecutter.package_url }}'),
                ('PyInventory', '{{ cookiecutter.package_name }}'),
            ),
        )

    def test_replace_path(self):
        path = replace_path(
            path=Path('foo', 'bar', 'baz'),
            reverse_info=(
                ('foo', '{{ package_name }}'),
                ('bar', '{{ dir_name }}'),
            ),
        )
        self.assertEqual(path, Path('{{ package_name }}/{{ dir_name }}/baz'))

    def test_build_dst_path(self):
        path = build_dst_path(
            source_path=Path('/the/source/path/'),
            item=Path('/the/source/path/foo/bar/test.py'),
            destination=Path('/the/destination'),
            reverse_info=(
                ('foo', '{{ package_name }}'),
                ('bar', '{{ dir_name }}'),
            ),
        )
        self.assertEqual(
            path,
            Path('/the/destination/{{ package_name }}/{{ dir_name }}/test.py'),
        )



import dataclasses
import datetime
import logging
import os
import subprocess  # nosec B404
from pathlib import Path
from shutil import which
from typing import Optional

from bx_py_utils.path import assert_is_file
from packaging.version import InvalidVersion, Version

from manageprojects.utilities.subprocess_utils import verbose_check_call, verbose_check_output


logger = logging.getLogger(__name__)


class GitError(BaseException):
    """Base class for all git errors"""


class NoGitRepoError(GitError):
    def __init__(self, path):
        super().__init__(f'"{path}" is not a git repository')


class GitBinNotFoundError(GitError):
    def __init__(self):
        super().__init__('Git executable not found in PATH')


@dataclasses.dataclass
class GitTagInfo:
    raw_tag: str
    version: Optional[Version] = None

    @property
    def version_tag(self):
        if self.version:
            return f'v{self.version}'
        return self.raw_tag


@dataclasses.dataclass
class GitTagInfos:
    tags: list[GitTagInfo]

    @classmethod
    def from_raw_tags(cls, raw_tags: list[str]) -> 'GitTagInfos':
        tags = []
        for raw_tag in raw_tags:
            try:
                version = Version(raw_tag)
            except InvalidVersion as err:
                logger.warning(f'Ignore: {err}')
                version = None
            tags.append(GitTagInfo(raw_tag=raw_tag, version=version))

        null_version = Version('0')
        tags.sort(key=lambda x: x.version or null_version)

        return cls(tags=tags)

    def get_releases(self):
        result = []
        for tag in self.tags:
            if version := tag.version:
                if version.is_devrelease or version.is_prerelease:
                    continue
                result.append(tag)
        return result

    def get_last_release(self):
        if releases := self.get_releases():
            return releases[-1]

    def exists(self, version: Version):
        for tag in self.tags:
            if tag.version and tag.version == version:
                return True
        return False


def get_git_root(path: Path):
    if len(path.parts) == 1 or not path.is_dir():
        return None

    if Path(path / '.git').is_dir():
        return path

    return get_git_root(path=path.parent)


class Git:
    def __init__(self, *, cwd: Path, detect_root: bool = True):
        if detect_root:
            self.cwd = get_git_root(cwd)
            if not self.cwd:
                raise NoGitRepoError(cwd)
        else:
            self.cwd = cwd

        self.git_bin = which('git')
        if not self.git_bin:
            raise GitBinNotFoundError()

        self.env = dict(os.environ)
        # no translated git command output ;)
        self.env['LANG'] = 'en_US.UTF-8'
        self.env['LANGUAGE'] = 'en_US'

    def git_verbose_check_call(self, *popenargs, **kwargs):
        popenargs = [self.git_bin, *popenargs]
        return verbose_check_call(
            *popenargs,
            cwd=self.cwd,
            env=self.env,
            **kwargs,
        )

    def git_verbose_check_output(self, *popenargs, **kwargs):
        popenargs = [self.git_bin, *popenargs]
        return verbose_check_output(
            *popenargs,
            cwd=self.cwd,
            env=self.env,
            **kwargs,
        )

    def git_verbose_output(
        self, *popenargs, exit_on_error=False, ignore_process_error=True, **kwargs
    ):
        popenargs = [self.git_bin, *popenargs]
        try:
            return verbose_check_output(
                *popenargs, cwd=self.cwd, exit_on_error=exit_on_error, **kwargs
            )
        except subprocess.CalledProcessError as err:
            if ignore_process_error:
                return err.stdout
            raise

    def get_current_hash(self, commit='HEAD', verbose=True) -> str:
        output = self.git_verbose_check_output('rev-parse', '--short', commit, verbose=verbose)
        if git_hash := output.strip():
            assert len(git_hash) == 7, f'No git hash from: {output!r}'
            return git_hash

        raise AssertionError(f'No git hash from: {output!r}')

    def get_commit_date(self, commit='HEAD', verbose=True) -> datetime.datetime:
        output = self.git_verbose_check_output(
            'show', '-s', '--format=%cI', commit, verbose=verbose
        )
        raw_date = output.strip()
        # e.g.: "2022-10-25 20:43:10 +0200"
        return datetime.datetime.fromisoformat(raw_date)

    def diff(
        self,
        reference1,
        reference2,
        no_color=True,
        indent_heuristic=False,
        irreversible_delete=True,
        verbose=True,
    ) -> str:
        # https://git-scm.com/docs/git-diff
        args = []
        if no_color:
            args.append('--no-color')

        if not indent_heuristic:
            args.append('--no-indent-heuristic')

        if irreversible_delete:
            args.append('--irreversible-delete')

        output = self.git_verbose_output('diff', *args, reference1, reference2, verbose=verbose)
        return output

    def apply(self, patch_path, verbose=True):
        # https://git-scm.com/docs/git-apply
        self.git_verbose_check_call(
            'apply',
            '--reject',
            '--ignore-whitespace',
            '--whitespace=fix',
            '-C1',
            '--recount',
            '--stat',
            '--summary',
            '--verbose',
            '--apply',
            patch_path,
            verbose=verbose,
        )

    def init(
        self,
        branch_name='main',
        user_name='manageprojects',
        user_email='manageprojects@test.tld',
        verbose=True,
    ) -> Path:
        output = self.git_verbose_check_output(
            'init', '-b', branch_name, verbose=verbose, exit_on_error=True
        )
        assert 'initialized' in output.lower(), f'Seems there is an error: {output}'
        self.cwd = get_git_root(self.cwd)

        self.config('user.name', user_name, scope='local', verbose=verbose)
        self.config('user.email', user_email, scope='local', verbose=verbose)

        return self.cwd

    def config(self, key, value, scope='local', verbose=True):
        assert scope in ('global', 'system', 'local')
        output = self.git_verbose_check_output(
            'config', f'--{scope}', key, value, verbose=verbose, exit_on_error=True
        )
        return output.strip()

    def get_config(self, key, verbose=True):
        if key in self.list_config_keys(verbose=verbose):
            output = self.git_verbose_check_output(
                'config', '--get', key, verbose=verbose, exit_on_error=False
            )
            return output.strip()

    def list_config_keys(self, verbose=False) -> set:
        output = self.git_verbose_check_output(
            'config', '--list', '--name-only', verbose=verbose, exit_on_error=False
        )
        keys = {item.strip() for item in output.splitlines() if item.strip()}
        return keys

    def add(self, spec, verbose=True) -> None:
        output = self.git_verbose_check_output('add', spec, verbose=verbose, exit_on_error=True)
        assert not output, f'Seems there is an error: {output}'

    def commit(self, comment, verbose=True) -> str:
        output = self.git_verbose_check_output(
            'commit', '--message', comment, verbose=verbose, exit_on_error=True
        )
        assert comment in output, f'Seems there is an error: {output}'
        return output

    def reflog(self, verbose=True) -> str:
        return self.git_verbose_check_output('reflog', verbose=verbose, exit_on_error=True)

    def log(
        self,
        format='%h - %an, %ar : %s',
        no_merges=False,
        commit1: Optional[str] = None,  # e.g.: "HEAD"
        commit2: Optional[str] = None,  # e.g.: "v0.8.0"
        verbose=True,
        exit_on_error=True,
    ) -> list[str]:
        """
        e.g.: git log --no-merges HEAD...v0.8.0 --pretty=format:"%h %as %s"
        """
        args = ['log']
        if no_merges:
            args.append('--no-merges')

        if commit1 and commit2:
            args.append(f'{commit1}...{commit2}')

        args.append(f'--pretty=format:{format}')
        output = self.git_verbose_check_output(*args, verbose=verbose, exit_on_error=exit_on_error)
        lines = output.splitlines()
        return lines

    def tag(self, git_tag: str, message: str, verbose=True, exit_on_error=True):
        self.git_verbose_check_call('tag', '-a', git_tag, '-m', message, verbose=verbose, exit_on_error=exit_on_error)

    def push(self, tags=False, verbose=True):
        if tags:
            args = ['--tags']
        else:
            args = []

        self.git_verbose_check_call('push', *args, verbose=verbose)

    def tag_list(self, verbose=True, exit_on_error=True) -> list[str]:
        output = self.git_verbose_check_output('tag', verbose=verbose, exit_on_error=exit_on_error)
        lines = output.splitlines()
        return lines

    def get_tag_infos(self, verbose=True, exit_on_error=True) -> GitTagInfos:
        raw_tags = self.tag_list(verbose=verbose, exit_on_error=exit_on_error)
        return GitTagInfos.from_raw_tags(raw_tags)

    def ls_files(self, verbose=True) -> list[Path]:
        output = self.git_verbose_check_output('ls-files', verbose=verbose, exit_on_error=True)
        file_paths = []
        for line in output.splitlines():
            file_path = self.cwd / line
            assert_is_file(file_path)
            file_paths.append(file_path)
        return sorted(file_paths)

    def print_file_list(self, out_func=print, verbose=True) -> None:
        for item in self.ls_files(verbose=verbose):
            out_func(f'* "{item.relative_to(self.cwd)}"')

    def reset(self, *, commit, hard=True, verbose=True) -> None:
        args = ['reset']
        if hard:
            args.append('--hard')
        args.append(commit)
        output = self.git_verbose_check_output(*args, verbose=verbose, exit_on_error=True)
        test_str = f'HEAD is now at {commit} '
        assert output.startswith(
            test_str
        ), f'Reset error: {output!r} does not start with {test_str!r}'

    def status(self, verbose=True) -> list:
        """
        Returns the changed files, if any.
        """
        output = self.git_verbose_check_output('status', '--porcelain', verbose=verbose, exit_on_error=True)
        result = []
        for line in output.splitlines():
            line = line.strip()
            status, filepath = line.split(' ', 1)
            result.append((status, filepath))
        return result

    def get_raw_branch_names(self, verbose=True) -> list[str]:
        output = self.git_verbose_check_output('branch', '--no-color', verbose=verbose)
        branches = output.splitlines()
        logger.debug('Git raw branches: %r', branches)
        return branches

    def get_current_branch_name(self, verbose=True) -> Optional[str]:
        raw_branch_names = self.get_raw_branch_names(verbose=verbose)
        for branch_name in raw_branch_names:
            if branch_name.startswith('*'):
                return branch_name.strip('* ')
        raise GitError(f'Current branch name not found in: {raw_branch_names}')

    def get_branch_names(self, verbose=True) -> list[str]:
        return sorted(branch.strip('* ') for branch in self.get_raw_branch_names(verbose=verbose))

    def get_main_branch_name(self, possible_names=('main', 'master'), verbose=True):
        """
        Returns the name of the "main" git branch.
        """
        branches = self.get_branch_names(verbose=verbose)

        branch_name = None
        for name in possible_names:
            if name in branches:
                branch_name = name
                break

        if branch_name is None:
            raise GitError(f'Git main branch not found in: {branches}')

        logger.info('Git main branch: "%s"', branch_name)
        return branch_name



from unittest.mock import patch

from click.testing import Result

from manageprojects.cli import cli_app
from manageprojects.cli.cli_app import cli
from manageprojects.test_utils.click_cli_utils import ClickInvokeCliException, invoke_click
from manageprojects.tests.base import BaseTestCase


class CliTestCase(BaseTestCase):
    def test_invoke_click(self):
        with patch.object(cli_app, 'verbose_check_call', side_effect=RuntimeError('Bam!')):
            with self.assertRaises(ClickInvokeCliException) as cm:
                invoke_click(cli, 'install')

        self.assertIsInstance(cm.exception, ClickInvokeCliException)
        self.assertIsInstance(cm.exception.result, Result)
        self.assertIsInstance(cm.exception.result.exception, RuntimeError)
        self.assertEqual(cm.exception.result.exception.args, ('Bam!',))



import atexit
import logging
import tempfile

from bx_py_utils.test_utils.log_utils import RaiseLogUsage


def logger_setup(*, logger_name, level, format, log_filename, raise_log_output):
    logger = logging.getLogger(logger_name)
    is_configured = logger.handlers and logger.level
    if not is_configured:
        logger.setLevel(level)

        if raise_log_output:
            handler = RaiseLogUsage()
        elif log_filename:
            handler = logging.FileHandler(filename=log_filename)
        else:
            handler = logging.StreamHandler()
        handler.setLevel(level)
        handler.setFormatter(logging.Formatter(format))

        logger.addHandler(handler)


def print_log_info(filename):
    print(f'\nLog file created here: {filename}\n')


def log_config(
    level=logging.DEBUG,
    format='%(asctime)s %(levelname)s %(name)s.%(funcName)s %(lineno)d | %(message)s',
    log_in_file=True,
    raise_log_output=False,
):
    if log_in_file:
        log_file = tempfile.NamedTemporaryFile(
            prefix='manageprojects_', suffix='.log', delete=False
        )
        log_filename = log_file.name
    else:
        log_filename = None

    logger_setup(
        logger_name='manageprojects',
        level=level,
        format=format,
        log_filename=log_filename,
        raise_log_output=raise_log_output,
    )
    logger_setup(
        logger_name='cookiecutter',
        level=level,
        format=format,
        log_filename=log_filename,
        raise_log_output=raise_log_output,
    )

    if log_filename:
        atexit.register(print_log_info, log_filename)


def log_func_call(*, logger, func, **kwargs):
    func_name = func.__name__
    logger.debug('Call %r with: %r', func_name, kwargs)
    result = func(**kwargs)
    logger.debug('%r result: %r', func_name, result)
    return result



import os
import unittest.util

from manageprojects.utilities.log_utils import log_config


# Hacky way to expand the failed test output:
unittest.util._MAX_LENGTH = os.environ.get('UNITTEST_MAX_LENGTH', 300)


log_config(
    format='%(levelname)s %(name)s.%(funcName)s %(lineno)d | %(message)s',
    log_in_file=False,
    raise_log_output=True,
)



import dataclasses
from pathlib import Path
from typing import Optional

from bx_py_utils.dict_utils import dict_get
from editorconfig import EditorConfigError, get_properties
from packaging.specifiers import SpecifierSet
from packaging.version import Version
from rich import print  # noqa
from rich.pretty import pprint

from manageprojects.constants import FORMAT_PY_FILE_DEFAULT_MAX_LINE_LENGTH, FORMAT_PY_FILE_DEFAULT_MIN_PYTON_VERSION
from manageprojects.exceptions import NoPyProjectTomlFound
from manageprojects.git import Git, GitError, NoGitRepoError
from manageprojects.utilities.pyproject_toml import TomlDocument, get_pyproject_toml
from manageprojects.utilities.subprocess_utils import ToolsExecutor


MAX_PY3_VER = 15


@dataclasses.dataclass
class GitInfo:
    cwd: Path
    main_branch_name: str


@dataclasses.dataclass
class PyProjectInfo:
    py_min_ver: Version
    pyproject_toml_path: Optional[Path] = None
    raw_py_ver_req: Optional[str] = None


@dataclasses.dataclass
class Config:
    git_info: Optional[GitInfo]
    pyproject_info: PyProjectInfo
    max_line_length: int

    @property
    def py_ver_str(self) -> Optional[str]:
        if pyproject_info := self.pyproject_info:
            return f'py{pyproject_info.py_min_ver.major}{pyproject_info.py_min_ver.minor}'

    @property
    def project_root_path(self) -> Optional[Path]:
        if git_info := self.git_info:
            if cwd := git_info.cwd:
                return cwd

        if pyproject_toml_path := self.pyproject_info.pyproject_toml_path:
            return pyproject_toml_path.parent

    @property
    def main_branch_name(self) -> Optional[str]:
        if git_info := self.git_info:
            return git_info.main_branch_name


def get_git_info(file_path: Path) -> Optional[GitInfo]:
    try:
        git = Git(cwd=file_path.parent, detect_root=True)
    except NoGitRepoError:
        print('File is not under Git version control')
    else:
        try:
            main_branch_name = git.get_main_branch_name(verbose=False)
        except GitError as err:
            print(f'Error: {err}')
        else:
            print(f'Git main branch name: {main_branch_name!r} from: {git.cwd}')
            return GitInfo(
                cwd=git.cwd,
                main_branch_name=main_branch_name,
            )


def get_py_min_version(raw_py_ver_req) -> Optional[Version]:
    specifier = SpecifierSet(raw_py_ver_req)

    # FIXME: How can this been simpler?
    for minor in range(MAX_PY3_VER):
        py_min_ver = f'3.{minor}'
        if specifier.contains(py_min_ver):
            py_min_ver = Version(py_min_ver)
            print(f'Min. Python version: {py_min_ver}')
            return py_min_ver

    print(f'Error getting min. python version from: {raw_py_ver_req!r} ({specifier})')


def get_pyproject_info(file_path: Path, default_min_py_version: str) -> PyProjectInfo:
    pyproject_info = PyProjectInfo(
        py_min_ver=Version(default_min_py_version),
    )

    try:
        toml_document: TomlDocument = get_pyproject_toml(file_path=file_path)
    except NoPyProjectTomlFound as err:
        print(err)
        print('Cannot detect the minimal Python version')
        return pyproject_info

    pyproject_info.pyproject_toml_path = toml_document.file_path
    data = toml_document.doc.unwrap()  # TOMLDocument -> dict

    if raw_py_ver_req := dict_get(data, 'project', 'requires-python'):
        # [project]
        # requires-python = ">=3.9,<4"
        pyproject_info.raw_py_ver_req = raw_py_ver_req
    elif raw_py_ver_req := dict_get(data, 'tool', 'poetry', 'dependencies', 'python'):
        # [tool.poetry.dependencies]
        # python = '>=3.9,<4.0.0'
        pyproject_info.raw_py_ver_req = raw_py_ver_req

    if pyproject_info.raw_py_ver_req:
        if py_min_ver := get_py_min_version(pyproject_info.raw_py_ver_req):
            pyproject_info.py_min_ver = py_min_ver

    return pyproject_info


def get_editorconfig_max_line_length(file_path) -> Optional[int]:
    try:
        options = get_properties(file_path)
    except EditorConfigError as err:
        print(f'EditorConfig error: {err}')
    else:
        max_line_length = options.get('max_line_length')
        if not max_line_length:
            print(f'No "max_line_length" in EditorConfig: {dict(options)}')
        else:
            max_line_length = int(max_line_length)
            print(f'Get max_line_length from EditorConfig: {max_line_length!r}')
            return max_line_length


def get_config(
    file_path,
    default_min_py_version=FORMAT_PY_FILE_DEFAULT_MIN_PYTON_VERSION,
    default_max_line_length=FORMAT_PY_FILE_DEFAULT_MAX_LINE_LENGTH,
) -> Config:
    config = Config(
        git_info=get_git_info(file_path),
        pyproject_info=get_pyproject_info(file_path, default_min_py_version),
        max_line_length=get_editorconfig_max_line_length(file_path) or default_max_line_length,
    )
    pprint(config)
    return config


def run_pyupgrade(tools_executor, file_path, config):
    pyver_arg = f'--{config.py_ver_str}-plus'
    tools_executor.verbose_check_output(
        'pyupgrade',
        '--exit-zero-even-if-changed',
        pyver_arg,
        file_path,
    )


def run_autoflake(tools_executor, file_path, config):
    # TODO: Remove if isort can do the job: https://github.com/PyCQA/isort/issues/1105
    tools_executor.verbose_check_output(
        'autoflake',
        '--in-place',
        file_path,
    )


def run_darker(tools_executor, file_path, config, darker_prefixes):
    if darker_prefixes:
        # darker/black will not fix e.g.:
        #   "E302 expected 2 blank lines"
        #   "E303 too many blank lines (4)"
        # work-a-round: run autopep8 only for these fixes ;)
        tools_executor.verbose_check_output(
            'autopep8',
            '--ignore-local-config',
            '--select',
            darker_prefixes,
            f'--max-line-length={config.max_line_length}',
            '--in-place',
            str(file_path),
        )

    tools_executor.verbose_check_output(
        'darker',
        '--flynt',
        '--isort',
        '--skip-string-normalization',
        '--revision',
        f'{config.main_branch_name}...',
        '--line-length',
        str(config.max_line_length),
        '--target-version',
        config.py_ver_str,
        file_path,
    )


def run_autopep8(tools_executor, file_path, config):
    tools_executor.verbose_check_output(
        'autopep8',
        '--ignore-local-config',
        '--max-line-length',
        str(config.max_line_length),
        '--aggressive',
        '--aggressive',
        '--in-place',
        file_path,
    )


def run_flake8(tools_executor, file_path, config):
    tools_executor.verbose_check_output(
        'flake8',
        '--max-line-length',
        str(config.max_line_length),
        file_path,
    )


def run_pyflakes(tools_executor, file_path, config):
    tools_executor.verbose_check_output('pyflakes', file_path)


def run_codespell(tools_executor, file_path, config):
    tools_executor.verbose_check_output('codespell', file_path)


def run_mypy(tools_executor, file_path, config):
    tools_executor.verbose_check_output(
        'mypy',
        '--ignore-missing-imports',
        '--follow-imports',
        'skip',
        '--allow-redefinition',  # https://github.com/python/mypy/issues/7165
        str(file_path),
    )


def format_one_file(
    *,
    default_min_py_version: str,
    default_max_line_length: int,
    darker_prefixes: str,
    file_path: Path,
) -> None:
    file_path = file_path.resolve()
    print(f'\nApply code formatter to: {file_path}')
    if file_path.suffix.lower() != '.py':
        print('Skip non-Python file ;)')
        return

    abs_file_path = file_path.absolute()

    old_content = abs_file_path.read_bytes()

    config = get_config(
        abs_file_path,
        default_min_py_version=default_min_py_version,
        default_max_line_length=default_max_line_length,
    )
    if cwd := config.project_root_path:
        file_path = abs_file_path.relative_to(cwd)

    tools_executor = ToolsExecutor(cwd=config.project_root_path)

    print('\n')

    run_pyupgrade(tools_executor, file_path, config)
    run_autoflake(tools_executor, file_path, config)

    if config.main_branch_name:
        run_darker(tools_executor, file_path, config, darker_prefixes)
    else:
        run_autopep8(tools_executor, file_path, config)

    run_flake8(tools_executor, file_path, config)
    run_pyflakes(tools_executor, file_path, config)
    run_codespell(tools_executor, file_path, config)
    run_mypy(tools_executor, file_path, config)

    print('\n')

    changed = old_content != abs_file_path.read_bytes()
    if changed:
        print(f'[green bold]*** File [blue]{abs_file_path}[/blue] successfully updated. ***')
    else:
        print(f'[green bold]*** File [blue]{abs_file_path}[/blue] needs to changes, ok. ***')



import logging
from pathlib import Path
from typing import Optional
from unittest.mock import patch

from bx_py_utils.path import assert_is_dir
from cookiecutter import generate
from cookiecutter.config import get_user_config
from cookiecutter.main import cookiecutter
from cookiecutter.repository import determine_repo_dir

from manageprojects.git import Git
from manageprojects.utilities.cookiecutter_utils import CookieCutterHookHandler
from manageprojects.utilities.log_utils import log_func_call


logger = logging.getLogger(__name__)


def get_repo_path(
    *,
    template: str,  # CookieCutter Template path or GitHub url
    directory: Optional[str] = None,  # Directory name of the CookieCutter Template
    checkout: Optional[str] = None,  # The branch, tag or commit ID to checkout after clone
    password: Optional[str] = None,  # Optional password to use when extracting the repository
    config_file: Optional[Path] = None,  # Optional path to 'cookiecutter_config.yaml'
) -> Path:
    """
    Checkout the cookiecutter template and reset it to `checkout` if needed.
    """
    if directory:
        assert '://' not in directory
        assert template != directory

    config_dict = log_func_call(
        logger=logger,
        func=get_user_config,
        config_file=config_file,
        default_config=None,
    )
    repo_dir, cleanup = determine_repo_dir(
        template=template,
        directory=directory,
        abbreviations=config_dict['abbreviations'],
        clone_to_dir=config_dict['cookiecutters_dir'],
        checkout=checkout,
        no_input=True,
        password=password,
    )
    logger.debug('repo_dir: %s', repo_dir)

    repo_path = Path(repo_dir)
    assert_is_dir(repo_path)

    if checkout is not None:
        # Cookiecutter will only checkout a specific commit, if `template` is a repro url!

        git = Git(cwd=repo_path, detect_root=True)
        assert repo_path.is_relative_to(git.cwd), f'Git repro {git.cwd} is not {repo_path}'
        # assert git.cwd == repo_path, f'Git repro {git.cwd} is not {repo_path}'
        current_hash = git.get_current_hash(verbose=True)
        if current_hash != checkout:
            git.reset(commit=checkout, verbose=True)

    return repo_path


def execute_cookiecutter(
    *,
    template: str,  # CookieCutter Template path or GitHub url
    directory: Optional[str] = None,  # Directory name of the CookieCutter Template
    output_dir: Path,  # Target path where CookieCutter should store the result files
    no_input: bool = False,  # Prompt the user at command line for manual configuration?
    extra_context: Optional[dict] = None,
    replay: Optional[bool] = None,
    checkout: Optional[str] = None,  # Optional branch, tag or commit ID to checkout after clone
    password: Optional[str] = None,  # Optional password to use when extracting the repository
    config_file: Optional[Path] = None,  # Optional path to 'cookiecutter_config.yaml'
) -> tuple[dict, Path, Path]:
    """
    "Just" run cookiecutter
    """
    repo_path = get_repo_path(
        template=template,
        directory=directory,
        checkout=checkout,
        password=password,
        config_file=config_file,
    )
    run_hook = CookieCutterHookHandler(origin_run_hook=generate.run_hook)
    with patch.object(generate, 'run_hook', run_hook):
        destination = log_func_call(
            logger=logger,
            func=cookiecutter,
            template=template,
            directory=directory,
            output_dir=output_dir,
            no_input=no_input,
            extra_context=extra_context,
            replay=replay,
            checkout=checkout,
            password=password,
            config_file=config_file,
        )
    cookiecutter_context = run_hook.context
    logger.info('Cookiecutter context: %r', cookiecutter_context)
    destination_path = Path(destination)
    assert_is_dir(destination_path)
    logger.info('Cookiecutter generated here: %r', destination_path)
    return cookiecutter_context, destination_path, repo_path



from pathlib import Path
from unittest import TestCase

from bx_py_utils.path import assert_is_file

from manageprojects.test_utils.temp_utils import TempContentFile


class TempUtilsTestCase(TestCase):
    def test_temp_content_file(self):
        with TempContentFile("Foo Bar") as file_path:
            assert isinstance(file_path, Path)
            assert_is_file(file_path)
            assert Path(file_path).read_text() == "Foo Bar"
        assert file_path.exists() is False

        with TempContentFile("", prefix="pre", suffix=".suf") as file_path:
            file_name = file_path.name
            assert file_name.startswith("pre")
            assert file_name.endswith(".suf")
            assert file_path.suffixes == [".suf"]



import configparser
import warnings
from pathlib import Path

from bx_py_utils.path import assert_is_dir, assert_is_file


EDITOR_CONFIG_DEFAULTS = {
    'foobar.py': {
        'indent_style': 'space',
        'indent_size': '4',
        'end_of_line': 'lf',
        'charset': 'utf-8',
    },
    'Makefile': {'indent_style': 'tab'},
}


def deep_check_max_line_length(data: dict, max_line_length: int, path: list):
    assert isinstance(data, dict)
    for key, value in data.items():
        if key in ('line_length', 'max_line_length'):
            assert (
                value == max_line_length
            ), f'{key!r} in {" / ".join(path)} is {value!r} but should be {max_line_length} !'

        if isinstance(value, dict):
            deep_check_max_line_length(value, max_line_length, path=[*path, key])


def check_project_max_line_length(package_root, max_line_length: int):
    assert_is_dir(package_root)

    assert isinstance(max_line_length, int)
    assert 70 < max_line_length < 200

    try:
        import tomllib  # New in Python 3.11
    except ImportError:
        try:
            import tomli as tomllib
        except ImportError as err:
            raise ImportError(f'Please add "tomli" to your dev-dependencies! Origin error: {err}')

    pyproject_toml = package_root / 'pyproject.toml'
    if pyproject_toml.is_file():
        pyproject_text = pyproject_toml.read_text(encoding="utf-8")
        pyproject_config = tomllib.loads(pyproject_text)
        deep_check_max_line_length(pyproject_config, max_line_length, path=['pyproject.toml'])


def _get_editor_config_options(file_path):
    try:
        from editorconfig import get_properties
    except ImportError as err:
        raise ImportError(f'Please add "EditorConfig" to your dev-dependencies! Origin error: {err}')

    options = get_properties(file_path)
    return options


def get_py_max_line_length(package_root) -> int:
    options = _get_editor_config_options(package_root / 'foobar.py')
    try:
        max_line_length = int(options['max_line_length'])
    except KeyError:
        raise KeyError('Editor config for Python files does not define "max_line_length" !')
    return max_line_length


def check_flake8_max_line_length(package_root, max_line_length):
    flake8_path = package_root / '.flake8'
    if not flake8_path.is_file():
        return

    config = configparser.RawConfigParser()
    config.read(flake8_path, encoding='UTF-8')

    try:
        flake8_max_line_length = config['flake8']['max-line-length']
    except KeyError as err:
        raise KeyError(f'Key "{err}" not found in Flake8 config file {flake8_path}')

    assert flake8_max_line_length == str(max_line_length), (
        f'Flake8 config file {flake8_path} as wrong max-line-length:'
        f' {flake8_max_line_length} but it should be: {max_line_length}'
    )


def check_editor_config(package_root: Path, config_defaults=None) -> None:
    """
    Check if "max line length" is the same in all config files, e.g.:
        - pyproject.toml
        - .editorconfig
        - .flake8
    """
    assert_is_dir(package_root)

    if config_defaults is None:
        config_defaults = EDITOR_CONFIG_DEFAULTS

    conf_filename = '.editorconfig'
    editor_config_file = package_root / conf_filename
    assert_is_file(editor_config_file)

    for example_filename, defaults in config_defaults.items():
        options = _get_editor_config_options(package_root / example_filename)
        for key, value in defaults.items():
            if key not in options:
                warnings.warn(
                    f'Editor config {key!r} for files like {example_filename} is not defined!',
                    RuntimeWarning,
                )
            else:
                current_value = options[key]
                if current_value != value:
                    warnings.warn(
                        (
                            f'Editor config {key!r} for files like {example_filename}'
                            f' should be {value!r} but is: {current_value!r}!'
                        ),
                        RuntimeWarning,
                    )

    max_line_length = get_py_max_line_length(package_root)
    check_project_max_line_length(package_root, max_line_length)
    check_flake8_max_line_length(package_root, max_line_length)



"""
    Allow manageprojects to be executable
    through `python -m manageprojects`.
"""


from manageprojects.cli import cli_app


def main():
    cli_app.main()


if __name__ == '__main__':
    main()



import sys
from pathlib import Path


from manageprojects.utilities.subprocess_utils import ToolsExecutor, verbose_check_call


def _call_darker(*args, package_root: Path, color: bool = True, verbose: bool = False):

    final_args = ['darker']
    if verbose:
        final_args.append('--verbose')
    if color:
        final_args.append('--color')

    final_args += list(args)

    tools_executor = ToolsExecutor(cwd=package_root)
    tools_executor.verbose_check_output(
        *final_args,
        exit_on_error=True,
    )


def fix(package_root: Path, color: bool = True, verbose: bool = False):
    """
    Fix code style via darker
    """
    _call_darker(color=color, verbose=verbose, package_root=package_root)
    print('Code style fixed, OK.')
    sys.exit(0)


def check(package_root: Path, color: bool = True, verbose: bool = False):
    """
    Check code style by calling darker + flake8
    """
    _call_darker('--check', color=color, verbose=verbose, package_root=package_root)

    if verbose:
        args = ['--verbose']
    else:
        args = []

    verbose_check_call(
        'flake8',
        *args,
        cwd=package_root,
        exit_on_error=True,
    )
    print('Code style: OK')
    sys.exit(0)



from bx_py_utils.auto_doc import assert_readme_block
from bx_py_utils.path import assert_is_file

from manageprojects import constants
from manageprojects.cli.cli_app import PACKAGE_ROOT, cli
from manageprojects.test_utils.click_cli_utils import invoke_click
from manageprojects.tests.base import BaseTestCase


def assert_cli_help_in_readme(text_block: str, marker: str):
    README_PATH = PACKAGE_ROOT / 'README.md'
    assert_is_file(README_PATH)

    text_block = text_block.replace(constants.CLI_EPILOG, '')
    text_block = f'```\n{text_block.strip()}\n```'
    assert_readme_block(
        readme_path=README_PATH,
        text_block=text_block,
        start_marker_line=f'[comment]: <> (✂✂✂ auto generated {marker} start ✂✂✂)',
        end_marker_line=f'[comment]: <> (✂✂✂ auto generated {marker} end ✂✂✂)',
    )


class ReadmeTestCase(BaseTestCase):
    def test_main_help(self):
        stdout = invoke_click(cli, '--help')
        self.assert_in_content(
            got=stdout,
            parts=(
                'Usage: ./cli.py [OPTIONS] COMMAND [ARGS]...',
                'start-project',
                'update-project',
                constants.CLI_EPILOG,
            ),
        )
        assert_cli_help_in_readme(text_block=stdout, marker='main help')

    def test_start_project_help(self):
        stdout = invoke_click(cli, 'start-project', '--help')
        self.assert_in_content(
            got=stdout,
            parts=(
                'Usage: ./cli.py start-project [OPTIONS] TEMPLATE OUTPUT_DIR',
                '--directory',
                '--input/--no-input',
            ),
        )
        assert_cli_help_in_readme(text_block=stdout, marker='start-project help')

    def test_update_project_help(self):
        stdout = invoke_click(cli, 'update-project', '--help')
        self.assert_in_content(
            got=stdout,
            parts=(
                'Usage: ./cli.py update-project [OPTIONS] PROJECT_PATH',
                '--input/--no-input',
                '--cleanup/--no-cleanup',
            ),
        )
        assert_cli_help_in_readme(text_block=stdout, marker='update-project help')

    def test_format_file_help(self):
        stdout = invoke_click(cli, 'format-file', '--help')
        self.assert_in_content(
            got=stdout,
            parts=(
                'Usage: ./cli.py format-file [OPTIONS] FILE_PATH',
                '--py-version',
                '--max-line-length',
                '--darker-prefixes',
            ),
        )
        assert_cli_help_in_readme(text_block=stdout, marker='format-file help')



import shutil
import tempfile
from collections.abc import Iterable
from importlib.metadata import version
from pathlib import Path
from typing import Optional

from bx_py_utils.dict_utils import dict_get
from bx_py_utils.path import assert_is_file

try:
    import tomllib  # New in Python 3.11
except ImportError:
    import tomli as tomllib

import sys

from packaging.version import Version
from rich import print  # noqa

from manageprojects.git import Git, GitError
from manageprojects.utilities.subprocess_utils import verbose_check_call, verbose_check_output


def exit_with_error(txt, hint=None):
    print(f'[red]ERROR: {txt}')
    if hint:
        print(f'[cyan](Hint: {hint})')
    print('[yellow]Abort.\n')
    sys.exit(-1)


def confirm(txt):
    print(f' [red]->[/red] [yellow]{txt}[/yellow]')
    print('>>>> [bold]Publish anyhow? (Y/N)', end=' ')
    if input().lower() not in ('y', 'j'):
        print('[cyan]Bye.\n')
        sys.exit(-1)


class PublisherGit:
    def __init__(
        self, *, package_path: Path, version: Version, possible_branch_names: Iterable[str], tag_msg_log_format: str
    ):
        self.package_path = package_path
        self.version = version
        self.possible_branch_names = possible_branch_names
        self.tag_msg_log_format = tag_msg_log_format

        self.git_tag_msg = None  # Set later

        print('\nDetect git repository...', end='')
        try:
            self.git = Git(cwd=package_path, detect_root=True)
        except GitError as err:
            exit_with_error(f'Git detection error: {err}')
            return

        if self.git.cwd != package_path:
            exit_with_error(f'Git path mismatch: {self.git.cwd} is not {package_path}')
        print('OK')

        print('\nAre we on the main branch...', end='')
        self.branch_name = self.git.get_current_branch_name(verbose=False)
        if self.branch_name not in possible_branch_names:
            confirm(f'The current git branch "[bold]{self.branch_name}[/bold]" seems to be wrong!')
        else:
            print('OK')

    def fast_checks(self):
        print('\nIs repository clean...', end='')
        if changes := self.git.status(verbose=False):
            confirm(f'Git changes: {changes}')
        else:
            print('OK')

        print(f'\ncheck git tag {self.version}...', end='')
        git_tag_infos = self.git.get_tag_infos()
        if git_tag_infos.exists(version=self.version):
            exit_with_error(f'git tag for "[bold]{self.version}[/bold]" already exists!')
        print('OK')

        print('\nCollect tag change message...', end='')
        last_tag = git_tag_infos.get_last_release()
        if not last_tag:
            # Just collect all commits
            commit1 = None
            commit2 = None
        else:
            commit1 = 'HEAD'
            commit2 = last_tag.raw_tag
            print(f'[blue]{commit1}...{commit2}[/blue]', end=' ')

        log_lines = self.git.log(
            format=self.tag_msg_log_format,
            no_merges=True,
            commit1=commit1,
            commit2=commit2,
            verbose=False,
            exit_on_error=True,
        )
        change_msg = '\n'.join(f' * {line}' for line in log_lines)
        print('done:')
        print(change_msg)
        if last_tag:
            self.git_tag_msg = (
                f'publish version v{self.version} with these changes (since {last_tag.version_tag}):\n\n{change_msg}\n'
            )
        else:
            self.git_tag_msg = f'publish version v{self.version} with these changes:\n\n{change_msg}\n'

    def slow_checks(self):
        print('\nLocal repository up-to-date: fetch origin...', end='')
        self.git.git_verbose_check_call('fetch', 'origin', verbose=False)
        print('compare...', end='')
        output = self.git.git_verbose_check_output(
            'log', f'HEAD..origin/{self.branch_name}', '--oneline', verbose=False
        )
        if output != '':
            exit_with_error(f'git repro is not up-to-date:\n{output}')
        else:
            print('push...', end='')
            output = self.git.git_verbose_check_output('push', 'origin', self.branch_name, verbose=False)
            if 'up-to-date' not in output:
                exit_with_error(f'git repro is not up-to-date:\n{output}')
        print('OK')

    def finalize(self):
        self.git.tag(f'v{self.version}', message=self.git_tag_msg, verbose=True)

        print('\ngit push tag to server')
        self.git.push(tags=True, verbose=True)


def clean_version(version: str) -> Version:
    assert version
    assert isinstance(version, str)
    return Version(version)


def get_pyproject_toml_version(package_path: Path) -> Optional[Version]:
    pyproject_toml_path = Path(package_path, 'pyproject.toml')
    assert_is_file(pyproject_toml_path)

    pyproject_toml = tomllib.loads(pyproject_toml_path.read_text(encoding='UTF-8'))

    ver_str = dict_get(pyproject_toml, 'project', 'version')
    if not ver_str:
        ver_str = dict_get(pyproject_toml, 'tool', 'poetry', 'version')

    if ver_str:
        return clean_version(ver_str)


def check_version(*, module, package_path: Path, distribution_name: Optional[str] = None) -> Version:
    if not distribution_name:
        distribution_name = module.__name__

    module_version = clean_version(module.__version__)
    installed_version = clean_version(version(distribution_name))
    if module_version != installed_version:
        exit_with_error(
            f'Version mismatch: current version {module_version} is not the installed one: {installed_version}',
            hint='Install package and run publish again',
        )

    pyproject_version = get_pyproject_toml_version(package_path)
    if not pyproject_version:
        confirm('Can not get package version from pyproject.toml')
    elif pyproject_version != installed_version:
        exit_with_error(
            (
                f'Version mismatch:'
                f' pyproject.toml version {pyproject_version} is not the installed one: {installed_version}'
            ),
            hint='Install package and run publish again',
        )

    return module_version


def build(package_path) -> None:
    print('\nCleanup old builds...', end='')

    def rmtree(path):
        if path.exists():
            print(f'remove old "{path.name}"...', end='')
            shutil.rmtree(path)

    rmtree(package_path / 'dist')
    rmtree(package_path / 'build')
    print('OK')

    if Path(package_path / 'poetry.lock').is_file():
        # Poetry used -> build with it
        output = verbose_check_output('poetry', 'build', verbose=True, exit_on_error=True)
    else:
        # use normal build
        output = verbose_check_output(sys.executable, '-m', 'build', verbose=True, exit_on_error=True)

    with tempfile.NamedTemporaryFile(
        mode='w', suffix='.txt', prefix=f'{package_path.name}_', encoding='utf-8', delete=False
    ) as temp_file:
        temp_file.write(output)
    print(f'[yellow]Output written to: [bold]{temp_file.name}')


def publish_package(
    *,
    module,
    package_path: Path,
    possible_branch_names: tuple[str] = ('main', 'master'),
    tag_msg_log_format: str = '%h %as %s',
    distribution_name: Optional[str] = None,  # Must be given, if it's not == module.__name__
) -> None:
    """
    Build and upload (with twine) a project to PyPi with many pre-checks:

     * Has correct version number?
     * Is on main branch and up-to-date with origin?
     * Check if current version already published
     * Build a git tag based on current package version
     * Adds change messages since last release to git tag message

    Designed to be useful for external packages.
    Some checks result in a hard exit, but some can be manually confirmed from the user to continue publishing.
    """
    # Version number correct?
    version: Version = check_version(module=module, package_path=package_path, distribution_name=distribution_name)
    if version.is_devrelease or version.is_prerelease:
        confirm(f'Current version {version} is dev/pre release!')

    pgit = PublisherGit(
        package_path=package_path,
        version=version,
        possible_branch_names=possible_branch_names,
        tag_msg_log_format=tag_msg_log_format,
    )
    pgit.fast_checks()

    # Build distribution from package:
    build(package_path)

    verbose_check_call('twine', 'check', '--strict', 'dist/*')

    # Are local git repository up-to-date with remote?
    pgit.slow_checks()

    verbose_check_call('twine', 'upload', 'dist/*')

    # Push the new tag to remote
    pgit.finalize()



import inspect
from pathlib import Path

from bx_py_utils.path import assert_is_file
from bx_py_utils.test_utils.datetime import parse_dt

import manageprojects
from manageprojects.data_classes import ManageProjectsMeta
from manageprojects.test_utils.logs import AssertLogs
from manageprojects.tests.base import BaseTestCase
from manageprojects.utilities.pyproject_toml import PyProjectToml, find_pyproject_toml
from manageprojects.utilities.temp_path import TemporaryDirectory


class PyProjectTomlTestCase(BaseTestCase):
    maxDiff = None

    def test_basic(self):
        with TemporaryDirectory(prefix='test_basic') as temp_path, AssertLogs(
            self, loggers=('manageprojects',)
        ) as logs:
            toml = PyProjectToml(project_path=temp_path)
            logs.assert_in('Create new pyproject.toml')

            self.assertEqual(toml.path, temp_path / 'pyproject.toml')
            self.assertFalse(toml.path.exists())

            toml.save()
            self.assert_file_content(
                toml.path,
                inspect.cleandoc(
                    '''
                    # Created by manageprojects

                    [manageprojects] # https://github.com/jedie/manageprojects
                    '''
                ),
            )
            self.assertEqual(
                toml.get_mp_meta(),
                ManageProjectsMeta(
                    initial_revision=None,
                    applied_migrations=[],
                    cookiecutter_template=None,
                    cookiecutter_directory=None,
                    cookiecutter_context=None,
                    initial_date=None,
                ),
            )

            toml.init(
                revision='abc0001',
                dt=parse_dt('2000-01-01T00:00:00+0000'),
                template='https://github.com/jedie/mp_test_template1/',
                directory='test_template1',
            )
            toml.save()
            self.assert_file_content(
                toml.path,
                inspect.cleandoc(
                    '''
                    # Created by manageprojects

                    [manageprojects] # https://github.com/jedie/manageprojects
                    initial_revision = "abc0001"
                    initial_date = 2000-01-01T00:00:00Z
                    cookiecutter_template = "https://github.com/jedie/mp_test_template1/"
                    cookiecutter_directory = "test_template1"
                    '''
                ),
            )

            toml.create_or_update_cookiecutter_context(
                context={
                    'cookiecutter': {
                        'foo': 'bar',
                        'x': 1,
                        '_template': '/foo/bar',
                    }
                }
            )
            toml.save()
            self.assert_file_content(
                toml.path,
                inspect.cleandoc(
                    '''
                    # Created by manageprojects

                    [manageprojects] # https://github.com/jedie/manageprojects
                    initial_revision = "abc0001"
                    initial_date = 2000-01-01T00:00:00Z
                    cookiecutter_template = "https://github.com/jedie/mp_test_template1/"
                    cookiecutter_directory = "test_template1"

                    [manageprojects.cookiecutter_context.cookiecutter]
                    foo = "bar"
                    x = 1
                    _template = "/foo/bar"
                    '''
                ),
            )

            # Reopen and get the information back:
            toml = PyProjectToml(project_path=temp_path)
            self.assertEqual(
                toml.get_mp_meta(),
                ManageProjectsMeta(
                    initial_revision='abc0001',
                    initial_date=parse_dt('2000-01-01T00:00:00+0000'),
                    applied_migrations=[],
                    cookiecutter_template='https://github.com/jedie/mp_test_template1/',
                    cookiecutter_directory='test_template1',
                    cookiecutter_context={
                        'cookiecutter': {
                            'foo': 'bar',
                            'x': 1,
                            '_template': '/foo/bar',
                        }
                    },
                ),
            )

            # add migration info
            toml.add_applied_migrations(git_hash='abc0002', dt=parse_dt('2000-02-02T00:00:00+0000'))
            toml.save()
            self.assert_file_content(
                toml.path,
                inspect.cleandoc(
                    '''
                    # Created by manageprojects

                    [manageprojects] # https://github.com/jedie/manageprojects
                    initial_revision = "abc0001"
                    initial_date = 2000-01-01T00:00:00Z
                    cookiecutter_template = "https://github.com/jedie/mp_test_template1/"
                    cookiecutter_directory = "test_template1"
                    applied_migrations = [
                        "abc0002", # 2000-02-02T00:00:00+00:00
                    ]

                    [manageprojects.cookiecutter_context.cookiecutter]
                    foo = "bar"
                    x = 1
                    _template = "/foo/bar"
                    '''
                ),
            )

            toml.add_applied_migrations(git_hash='abc0003', dt=parse_dt('2000-03-03T00:00:00+0000'))
            toml.create_or_update_cookiecutter_context(
                context={
                    'cookiecutter': {
                        'foo': 'overwritten',
                        'x': 2,
                        'new': 'value',
                        '_template': '/foo/bar',
                    }
                }
            )
            toml.save()
            self.assert_file_content(
                toml.path,
                inspect.cleandoc(
                    '''
                    # Created by manageprojects

                    [manageprojects] # https://github.com/jedie/manageprojects
                    initial_revision = "abc0001"
                    initial_date = 2000-01-01T00:00:00Z
                    cookiecutter_template = "https://github.com/jedie/mp_test_template1/"
                    cookiecutter_directory = "test_template1"
                    applied_migrations = [
                        "abc0002", # 2000-02-02T00:00:00+00:00
                        "abc0003", # 2000-03-03T00:00:00+00:00
                    ]

                    [manageprojects.cookiecutter_context.cookiecutter]
                    foo = "overwritten"
                    x = 2
                    new = "value"
                    _template = "/foo/bar"
                    '''
                ),
            )
            data = toml.get_mp_meta()
            self.assertEqual(
                data,
                ManageProjectsMeta(
                    initial_revision='abc0001',
                    initial_date=parse_dt('2000-01-01T00:00:00+0000'),
                    applied_migrations=['abc0002', 'abc0003'],
                    cookiecutter_template='https://github.com/jedie/mp_test_template1/',
                    cookiecutter_directory='test_template1',
                    cookiecutter_context={
                        'cookiecutter': {
                            'foo': 'overwritten',
                            'x': 2,
                            'new': 'value',
                            '_template': '/foo/bar',
                        }
                    },
                ),
            )
            self.assertIsInstance(data.cookiecutter_context, dict)
            self.assertEqual(type(data.cookiecutter_context), dict)
        logs.assert_in('Create new pyproject.toml', 'Read existing pyproject.toml')

    def test_expand_existing_toml(self):
        with TemporaryDirectory(prefix='test_basic') as temp_path, AssertLogs(
            self, loggers=('manageprojects',)
        ) as logs:
            file_path = temp_path / 'pyproject.toml'
            file_path.write_text(
                inspect.cleandoc(
                    '''
                    [project]
                    name = "manageprojects"
                    version = "0.0.2rc1"

                    [project.urls]
                    "Homepage" = "https://github.com/jedie/manageproject"
                    "Bug Tracker" = "https://github.com/jedie/manageproject/issues"
                    '''
                )
            )

            with AssertLogs(self, loggers=('manageprojects',)) as logs:
                toml = PyProjectToml(project_path=temp_path)
                toml.init(
                    revision='abc0001',
                    dt=parse_dt('2000-01-01T00:00:00+0000'),
                    template='/foo/bar/template',
                    directory=None,
                )
                toml.save()
            logs.assert_in('Read existing pyproject.toml')

            self.assert_file_content(
                file_path,
                inspect.cleandoc(
                    '''
                    [project]
                    name = "manageprojects"
                    version = "0.0.2rc1"

                    [project.urls]
                    "Homepage" = "https://github.com/jedie/manageproject"
                    "Bug Tracker" = "https://github.com/jedie/manageproject/issues"

                    [manageprojects] # https://github.com/jedie/manageprojects
                    initial_revision = "abc0001"
                    initial_date = 2000-01-01T00:00:00Z
                    cookiecutter_template = "/foo/bar/template"
                    '''
                ),
            )
            self.assertEqual(
                toml.get_mp_meta(),
                ManageProjectsMeta(
                    initial_revision='abc0001',
                    initial_date=parse_dt('2000-01-01T00:00:00+0000'),
                    applied_migrations=[],
                    cookiecutter_template='/foo/bar/template',
                    cookiecutter_directory=None,
                    cookiecutter_context=None,
                ),
            )
        logs.assert_in('Read existing pyproject.toml')

    def test_find_pyproject_toml(self):
        own_pyproject_toml = Path(manageprojects.__file__).parent.parent / 'pyproject.toml'
        assert_is_file(own_pyproject_toml)
        self.assertEqual(find_pyproject_toml(file_path=Path(__file__)), own_pyproject_toml)

        with TemporaryDirectory(prefix='test_find_pyproject_toml') as temp_path:
            self.assertIsNone(find_pyproject_toml(file_path=temp_path))



import logging


logger = logging.getLogger(__name__)


class CookieCutterHookHandler:
    """
    Capture the effective Cookiecutter Template Context via injecting the Cookiecutter hooks.
    """

    def __init__(self, origin_run_hook):
        self.origin_run_hook = origin_run_hook
        self.context = {}

    def __call__(self, hook_name, project_dir, context):
        logger.debug('Hook %r for %r context: %r', hook_name, project_dir, context)
        origin_hook_result = self.origin_run_hook(hook_name, project_dir, context)
        self.context.update(context)
        return origin_hook_result



from pathlib import Path
from unittest import TestCase

from bx_py_utils.test_utils.redirect import RedirectOut

import manageprojects
from manageprojects import __version__
from manageprojects.cli.cli_app import PACKAGE_ROOT
from manageprojects.git import Git
from manageprojects.test_utils.logs import AssertLogs
from manageprojects.utilities.temp_path import TemporaryDirectory
from manageprojects.utilities.version_info import print_version


class VersionInfoTestCase(TestCase):
    maxDiff = None

    def test_temp_content_file(self):
        git = Git(cwd=PACKAGE_ROOT)
        git_hash = git.get_current_hash(verbose=False)

        with RedirectOut() as buffer:
            print_version(module=manageprojects)

        self.assertEqual(buffer.stderr, '')
        self.assertEqual(buffer.stdout, f'manageprojects v{__version__} {git_hash} ({PACKAGE_ROOT})\n')

    def test_no_git(self):
        with RedirectOut() as buffer:
            print_version(module=manageprojects, project_root=Path('foo', 'bar'))

        self.assertEqual(buffer.stderr, '')
        self.assertEqual(buffer.stdout, f'manageprojects v{__version__} (No git found for: foo/bar)\n')

        with TemporaryDirectory(prefix='test_no_git') as temp_path:
            non_git_path = temp_path / '.git'
            non_git_path.mkdir()

            with AssertLogs(self, loggers=('manageprojects',)) as logs, RedirectOut() as buffer:
                print_version(module=manageprojects, project_root=temp_path)

            self.assertEqual(buffer.stderr, '')
            self.assertIn(f'manageprojects v{__version__} ', buffer.stdout)

            logs.assert_in('Error print version', 'Traceback')



import filecmp
import logging
import shutil
import sys
from pathlib import Path
from typing import Optional

from bx_py_utils.path import assert_is_dir
from rich import print

from manageprojects.cookiecutter_api import execute_cookiecutter
from manageprojects.data_classes import OverwriteResult
from manageprojects.git import Git
from manageprojects.utilities.temp_path import TemporaryDirectory


logger = logging.getLogger(__name__)


def overwrite_project(
    *,
    git: Git,
    project_path: Path,
    template: str,  # CookieCutter Template path or GitHub url
    from_rev: str,
    replay_context: dict,
    directory: Optional[str] = None,  # Directory name of the CookieCutter Template
    password: Optional[str] = None,
    config_file: Optional[Path] = None,  # Optional path to 'cookiecutter_config.yaml'
    cleanup: bool = True,  # Remove temp files if not exceptions happens
    no_input: bool = False,  # Prompt the user at command line for manual configuration?
) -> OverwriteResult:
    print(f'Update by overwrite project: {project_path} from {template}')

    status = git.status()
    if status:
        print(f'Abort: project {project_path} is not clean:', file=sys.stderr)
        for flag, filepath in status:
            print(flag, filepath)
        sys.exit(1)

    extra_context = replay_context.get('cookiecutter')
    if not extra_context:
        print('WARNING: No "cookiecutter" in replay context!')

    project_name = project_path.name
    with TemporaryDirectory(prefix=f'manageprojects_{project_name}_', cleanup=cleanup) as temp_path:
        print(f'Compile cookiecutter template in the current version here: {temp_path}')
        print('Use extra context:')
        print(extra_context)
        to_rev_context, to_rev_dst_path, to_rev_repo_path = execute_cookiecutter(
            template=template,
            directory=directory,
            output_dir=temp_path,
            no_input=no_input,
            extra_context=extra_context,
            checkout=None,  # Checkout HEAD/main revision
            password=password,
            config_file=config_file,
        )
        assert_is_dir(to_rev_repo_path)

        git = Git(cwd=to_rev_repo_path, detect_root=True)
        to_rev = git.get_current_hash(verbose=False)
        to_commit_date = git.get_commit_date(verbose=False)
        print(f'Update from rev. {from_rev} to rev. {to_rev} ({to_commit_date})')

        updated_file_count = 0
        for src_file_path in to_rev_dst_path.rglob('*'):
            if not src_file_path.is_file():
                continue

            dst_file_path = project_path / src_file_path.relative_to(to_rev_dst_path)
            if not dst_file_path.exists():
                print(f'NEW file: {dst_file_path}')
            elif filecmp.cmp(dst_file_path, src_file_path, shallow=False):
                print(f'Skip unchanged file: {dst_file_path}, ok.')
                continue
            else:
                print(f'UPDATE file: {dst_file_path}')

            if not dst_file_path.parent.exists():
                dst_file_path.parent.mkdir(parents=True)

            shutil.copyfile(src_file_path, dst_file_path)
            updated_file_count += 1

    logger.info('%i files updated by overwriting', updated_file_count)

    if updated_file_count > 0:
        return OverwriteResult(to_rev=to_rev, to_commit_date=to_commit_date)



import inspect
from pathlib import Path
from unittest import TestCase

from packaging.version import Version

from manageprojects.constants import PY_BIN_PATH
from manageprojects.format_file import (
    Config,
    GitInfo,
    PyProjectInfo,
    format_one_file,
    get_config,
    get_editorconfig_max_line_length,
    get_git_info,
    get_pyproject_info,
)
from manageprojects.test_utils.logs import AssertLogs
from manageprojects.test_utils.subprocess import SimpleRunReturnCallback, SubprocessCallMock
from manageprojects.tests.base import GIT_BIN_PARENT, PROJECT_PATH
from manageprojects.utilities.temp_path import TemporaryDirectory


class FormatFileTestCase(TestCase):
    def test_get_git_info(self):
        with AssertLogs(self, loggers=('manageprojects',)):
            git_info = get_git_info(file_path=Path(__file__))
            self.assertIsInstance(git_info, GitInfo)
            self.assertEqual(git_info.main_branch_name, 'main')

            with TemporaryDirectory(prefix='test_get_main_branch_name') as temp_path:
                self.assertIsNone(get_git_info(file_path=temp_path))

    def test_get_pyproject_info(self):
        with TemporaryDirectory(prefix='test_get_pyproject_info') as temp_path:
            self.assertEqual(
                get_pyproject_info(temp_path, default_min_py_version='1.2.3'),
                PyProjectInfo(
                    pyproject_toml_path=None,
                    raw_py_ver_req=None,
                    py_min_ver=Version('1.2.3'),
                ),
            )

            pyproject_toml = Path(temp_path, 'pyproject.toml')
            pyproject_toml.touch()
            self.assertEqual(
                get_pyproject_info(temp_path, default_min_py_version='3.8'),
                PyProjectInfo(
                    pyproject_toml_path=pyproject_toml,
                    raw_py_ver_req=None,
                    py_min_ver=Version('3.8'),
                ),
            )

            pyproject_toml.write_text(
                inspect.cleandoc(
                    '''
                    [project]
                    foo=1
                    requires-python = ">=3.10,<4"
                    bar=2
                    '''
                )
            )
            self.assertEqual(
                get_pyproject_info(temp_path, default_min_py_version='3.11'),
                PyProjectInfo(
                    pyproject_toml_path=pyproject_toml,
                    raw_py_ver_req='>=3.10,<4',
                    py_min_ver=Version('3.10'),
                ),
            )

            pyproject_toml.write_text(
                inspect.cleandoc(
                    '''
                    [tool.poetry.dependencies]
                    foo=1
                    python = '>=3.11,<4.0.0'
                    bar=2
                    '''
                )
            )
            self.assertEqual(
                get_pyproject_info(temp_path, default_min_py_version='3.8'),
                PyProjectInfo(
                    pyproject_toml_path=pyproject_toml,
                    raw_py_ver_req='>=3.11,<4.0.0',
                    py_min_ver=Version('3.11'),
                ),
            )

            pyproject_toml.write_text(
                inspect.cleandoc(
                    '''
                    [project]
                    requires-python = ">=3.99,<4"
                    '''
                )
            )
            self.assertEqual(
                get_pyproject_info(temp_path, default_min_py_version='3.11'),
                PyProjectInfo(
                    pyproject_toml_path=pyproject_toml,
                    raw_py_ver_req='>=3.99,<4',
                    py_min_ver=Version('3.11'),
                ),
            )

        self.assertEqual(
            get_pyproject_info(file_path=Path(__file__), default_min_py_version='3.7'),
            PyProjectInfo(
                pyproject_toml_path=PROJECT_PATH / 'pyproject.toml',
                py_min_ver=Version('3.9'),
                raw_py_ver_req='>=3.9,<4',
            ),
        )

    def test_get_editorconfig_max_line_length(self):
        self.assertEqual(get_editorconfig_max_line_length(file_path=Path(__file__)), 119)
        with TemporaryDirectory(prefix='test_get_main_branch_name') as temp_path:
            self.assertIsNone(get_editorconfig_max_line_length(file_path=temp_path))

    def test_get_config(self):
        with AssertLogs(self, loggers=('manageprojects',)):
            config = get_config(file_path=Path(__file__))
            self.assertIsInstance(config, Config)
            self.assertEqual(
                config,
                Config(
                    git_info=GitInfo(cwd=PROJECT_PATH, main_branch_name='main'),
                    pyproject_info=PyProjectInfo(
                        pyproject_toml_path=PROJECT_PATH / 'pyproject.toml',
                        py_min_ver=Version('3.9'),
                        raw_py_ver_req='>=3.9,<4',
                    ),
                    max_line_length=119,
                ),
            )
            with TemporaryDirectory(prefix='test_get_config') as temp_path:
                config = get_config(file_path=temp_path)
                self.assertIsInstance(config, Config)
                self.assertEqual(
                    config,
                    Config(
                        git_info=None,
                        pyproject_info=PyProjectInfo(
                            py_min_ver=Version('3.9'), pyproject_toml_path=None, raw_py_ver_req=None
                        ),
                        max_line_length=119,
                    ),
                )

    def test_format_one_file(self):
        with AssertLogs(self, loggers=('manageprojects',)), SubprocessCallMock(
            return_callback=SimpleRunReturnCallback(
                stdout='* main',  # "git branch" call -> bach name found -> use Darker
            )
        ) as call_mock:
            format_one_file(
                default_min_py_version='3.11',
                default_max_line_length=123,
                darker_prefixes='E456,E789',
                file_path=Path(__file__),
            )

        self.assertEqual(
            call_mock.get_popenargs(rstrip_paths=(PY_BIN_PATH, GIT_BIN_PARENT)),
            [
                ['.../git', 'branch', '--no-color'],
                [
                    '.../pyupgrade',
                    '--exit-zero-even-if-changed',
                    '--py39-plus',
                    'manageprojects/tests/test_format_file.py',
                ],
                ['.../autoflake', '--in-place', 'manageprojects/tests/test_format_file.py'],
                [
                    '.../autopep8',
                    '--ignore-local-config',
                    '--select',
                    'E456,E789',
                    '--max-line-length=119',
                    '--in-place',
                    'manageprojects/tests/test_format_file.py',
                ],
                [
                    '.../darker',
                    '--flynt',
                    '--isort',
                    '--skip-string-normalization',
                    '--revision',
                    'main...',
                    '--line-length',
                    '119',
                    '--target-version',
                    'py39',
                    'manageprojects/tests/test_format_file.py',
                ],
                ['.../flake8', '--max-line-length', '119', 'manageprojects/tests/test_format_file.py'],
                ['.../pyflakes', 'manageprojects/tests/test_format_file.py'],
                ['.../codespell', 'manageprojects/tests/test_format_file.py'],
                [
                    '.../mypy',
                    '--ignore-missing-imports',
                    '--follow-imports',
                    'skip',
                    '--allow-redefinition',
                    'manageprojects/tests/test_format_file.py',
                ],
            ],
        )

        with AssertLogs(self, loggers=('manageprojects',)), SubprocessCallMock(
            return_callback=SimpleRunReturnCallback(
                stdout='',  # No git branch name found -> Don't use Darker -> use autopep8
            )
        ) as call_mock:
            format_one_file(
                default_min_py_version='3.11',
                default_max_line_length=123,
                darker_prefixes='E456,E789',
                file_path=Path(__file__),
            )

        self.assertEqual(
            call_mock.get_popenargs(rstrip_paths=(PY_BIN_PATH, GIT_BIN_PARENT)),
            [
                ['.../git', 'branch', '--no-color'],
                [
                    '.../pyupgrade',
                    '--exit-zero-even-if-changed',
                    '--py39-plus',
                    'manageprojects/tests/test_format_file.py',
                ],
                ['.../autoflake', '--in-place', 'manageprojects/tests/test_format_file.py'],
                [
                    '.../autopep8',  # <<< Instead of Darker!
                    '--ignore-local-config',
                    '--max-line-length',
                    '119',
                    '--aggressive',
                    '--aggressive',
                    '--in-place',
                    'manageprojects/tests/test_format_file.py',
                ],
                ['.../flake8', '--max-line-length', '119', 'manageprojects/tests/test_format_file.py'],
                ['.../pyflakes', 'manageprojects/tests/test_format_file.py'],
                ['.../codespell', 'manageprojects/tests/test_format_file.py'],
                [
                    '.../mypy',
                    '--ignore-missing-imports',
                    '--follow-imports',
                    'skip',
                    '--allow-redefinition',
                    'manageprojects/tests/test_format_file.py',
                ],
            ],
        )



class ProjectNotFound(FileNotFoundError):
    pass


class NoManageprojectsMeta(Exception):
    """
    A existing 'pyproject.toml' file does not contain [manageprojects] information
    """

    pass


class NoPyProjectTomlFound(FileNotFoundError):
    pass



import os
import shutil
import subprocess
import sys
from pathlib import Path

from bx_py_utils.path import assert_is_dir, assert_is_file
from rich import print  # noqa

from manageprojects.constants import PY_BIN_PATH


DEFAULT_TIMEOUT = 5 * 60


def make_absolute_path(path):
    """
    Resolve the path, but not for symlinks.
    e.g.: '.venv/bin/python' is a symlink to e.g.: '/usr/bin/python3.x'
    We would like to keep the origin symlink ;)
    """
    if not isinstance(path, Path):
        path = Path(path)

    if path.is_symlink():
        path = path.absolute()
    else:
        path = path.resolve(strict=True)

    return path


def make_relative_path(path, relative_to):
    """
    Makes {path} relative to {relative_to}, if possible.
    """
    path = make_absolute_path(path=path)
    relative_to = make_absolute_path(path=relative_to)

    try:
        is_relative_to = path.is_relative_to(relative_to)
    except AttributeError:  # is_relative_to() is new in Python 3.9
        try:
            path = path.relative_to(relative_to)
        except ValueError:
            # {path} doesn't start with {relative_to} -> do nothing
            pass
    else:
        if is_relative_to:
            path = path.relative_to(relative_to)

    return path


def _print_info(popenargs, *, cwd, kwargs):
    print()
    print('_' * 100)

    if len(popenargs) > 1:
        command, *args = popenargs
    else:
        command = popenargs[0]
        args = []

    command_path = make_relative_path(Path(command), relative_to=cwd)

    command_name = command_path.name
    command_dir = command_path.parent

    print(f'[white]{cwd}[/white][bold]$[/bold]', end=' ')

    if command_dir and command_dir != Path.cwd():
        print(f'[green]{command_dir}{os.sep}[/green]', end='')

    if command_name:
        print(f'[yellow bold]{command_name}[/yellow bold]', end='')

    for arg in args:
        if isinstance(arg, str):
            if arg.startswith('--'):
                print(f' [magenta]{arg}[/magenta]', end='')
                continue
            elif arg.startswith('-'):
                print(f' [cyan]{arg}[/cyan]', end='')
                continue

        print(f' [blue]{arg}[/blue]', end='')

    if kwargs:
        verbose_kwargs = ', '.join(f'{k}={v!r}' for k, v in sorted(kwargs.items()))
        print(f' (kwargs: {verbose_kwargs})', end='')

    print('\n', flush=True)


def prepare_popenargs(popenargs, cwd=None):
    if cwd is None:
        cwd = Path.cwd()
    else:
        assert_is_dir(cwd)

    command_path = Path(popenargs[0])
    if not command_path.is_file():
        # Lookup in current venv bin path first:
        command = shutil.which(str(command_path), path=str(PY_BIN_PATH))
        if not command:
            # Search in PATH for this command that doesn't point to a existing file:
            command = shutil.which(str(command_path))
            if not command:
                raise FileNotFoundError(f'Command "{popenargs[0]}" not found in PATH!')

        command = make_absolute_path(path=command)

        # Replace command name with full path:
        popenargs = list(popenargs)
        popenargs[0] = command

    return popenargs, cwd


def verbose_check_call(
    *popenargs,
    verbose=True,
    cwd=None,
    extra_env=None,
    exit_on_error=False,
    timeout=DEFAULT_TIMEOUT,
    env=None,
    text=True,
    **kwargs,
):
    """
    'verbose' version of subprocess.check_call()
    Will try to find the command in current Python venv/bin/ path
    """

    popenargs, cwd = prepare_popenargs(popenargs, cwd=cwd)

    if verbose:
        _print_info(popenargs, cwd=cwd, kwargs=kwargs)

    if env is None:
        env = os.environ.copy()

    if extra_env:
        env.update(extra_env)

    try:
        return subprocess.check_call(
            [str(part) for part in popenargs],  # e.g.: Path() instance -> str,
            text=text,
            env=env,
            cwd=cwd,
            timeout=timeout,
            **kwargs,
        )
    except subprocess.CalledProcessError as err:
        if verbose:
            print(f'[red]Process "{popenargs[0]}" finished with exit code {err.returncode!r}[/red]')
        if exit_on_error:
            sys.exit(err.returncode)
        raise


def verbose_check_output(
    *popenargs,
    verbose=True,
    cwd=None,
    extra_env=None,
    exit_on_error=False,
    print_output_on_error=True,
    timeout=DEFAULT_TIMEOUT,
    env=None,
    text=True,
    **kwargs,
):
    """
    'verbose' version of subprocess.check_output()
    Will try to find the command in current Python venv/bin/ path
    """

    popenargs, cwd = prepare_popenargs(popenargs, cwd=cwd)

    if verbose:
        _print_info(popenargs, cwd=cwd, kwargs=kwargs)

    if env is None:
        env = os.environ.copy()

    if extra_env:
        env.update(extra_env)

    try:
        output = subprocess.check_output(
            [str(part) for part in popenargs],  # e.g.: Path() instance -> str,
            text=text,
            env=env,
            cwd=cwd,
            stderr=subprocess.STDOUT,
            timeout=timeout,
            **kwargs,
        )
    except subprocess.CalledProcessError as err:
        if print_output_on_error or exit_on_error:
            if not verbose:
                _print_info(popenargs, cwd=cwd, kwargs=kwargs)
            print(f'[red]Process "{popenargs[0]}" finished with exit code {err.returncode!r}[/red]')
            print('-' * 79)
            print(err.stdout)
            print('-' * 79)
        if exit_on_error:
            sys.exit(err.returncode)
        raise
    else:
        return output


class ToolsExecutor:
    def __init__(self, cwd: Path):
        self.cwd = cwd

        self.extra_env = {}

        bin_path_str = str(PY_BIN_PATH)
        if bin_path_str not in os.environ['PATH']:
            self.extra_env['PATH'] = bin_path_str + os.pathsep + os.environ['PATH']

        self.extra_env['PYTHONUNBUFFERED'] = '1'

    def verbose_check_output(self, bin, *popenargs, verbose=True, exit_on_error=True, **kwargs):
        """'verbose' version of subprocess.check_output()"""
        bin_path = PY_BIN_PATH / bin
        assert_is_file(bin_path)

        try:
            verbose_check_call(
                bin_path,
                *popenargs,
                verbose=verbose,
                cwd=self.cwd,
                extra_env=self.extra_env,
                exit_on_error=exit_on_error,
                **kwargs,
            )
        except subprocess.CalledProcessError:
            pass



import inspect
import json
from pathlib import Path

from bx_py_utils.path import assert_is_file
from bx_py_utils.test_utils.redirect import RedirectOut

from manageprojects.cookiecutter_templates import update_managed_project
from manageprojects.data_classes import ManageProjectsMeta, OverwriteResult
from manageprojects.test_utils.git_utils import init_git
from manageprojects.test_utils.logs import AssertLogs
from manageprojects.tests.base import BaseTestCase
from manageprojects.utilities.pyproject_toml import PyProjectToml
from manageprojects.utilities.temp_path import TemporaryDirectory


class UpdateByOverwriteTestCase(BaseTestCase):
    def test_happy_path(self):
        cookiecutter_context = {
            'dir_name': 'a_directory',
            'file_name': 'a_file_name',
            'value': 'FooBar',
        }
        context = {'cookiecutter': cookiecutter_context}

        with TemporaryDirectory(prefix='test_update_by_overwrite_') as main_temp_path:
            project_path = main_temp_path / 'project_path'
            dst_file_path = project_path / 'a_file_name.py'

            template_path = main_temp_path / 'template'
            template_dir_name = 'template_dir'
            template_dir_path = template_path / template_dir_name
            config_file_path = template_dir_path / 'cookiecutter.json'

            test_file_path = template_dir_path / '{{cookiecutter.dir_name}}' / '{{cookiecutter.file_name}}.py'

            dst_file_path.parent.mkdir(parents=True)
            dst_file_path.write_text('# This is a test line, not changed')

            config_file_path.parent.mkdir(parents=True)
            config_file_path.write_text(json.dumps(cookiecutter_context))

            project_git, project_from_rev = init_git(project_path, comment='Git init project.')
            dst_file_path.write_text(
                inspect.cleandoc(
                    '''
                    # This is a test line, not changed
                    #
                    # Revision 1
                    #
                    # The same cookiecutter value:
                    print('Test: FooBar')
                    '''
                )
            )
            project_git.add('.')
            project_git.commit('Store revision 1')

            test_file_path.parent.mkdir(parents=True)
            test_file_path.write_text(
                inspect.cleandoc(
                    '''
                    # This is a test line, not changed
                    #
                    # Revision 1
                    #
                    # The same cookiecutter value:
                    print('Test: {{ cookiecutter.value }}')
                    '''
                )
            )

            git, from_rev = init_git(template_path, comment='Git init template.')
            from_date = git.get_commit_date(verbose=False)

            with AssertLogs(self, loggers=('manageprojects',)) as logs:
                toml = PyProjectToml(project_path=project_path)
                toml.init(
                    revision=from_rev,
                    dt=from_date,
                    template=str(template_path),
                    directory=template_dir_name,
                )
            logs.assert_in('Create new pyproject.toml')
            toml.create_or_update_cookiecutter_context(context=context)
            toml.save()
            toml_content = toml.path.read_text()
            self.assertIn('# Created by manageprojects', toml_content)
            self.assertIn('[manageprojects] # https://github.com/jedie/manageprojects', toml_content)
            self.assertIn(f'initial_revision = "{from_rev}"', toml_content)
            self.assertIn(f'cookiecutter_template = "{template_path}"', toml_content)
            self.assertIn('cookiecutter_directory = "template_dir"', toml_content)

            test_file_path.write_text(
                inspect.cleandoc(
                    '''
                    # This is a test line, not changed
                    #
                    # Revision 2
                    #
                    # The same cookiecutter value:
                    print('Test: {{ cookiecutter.value }}')
                    '''
                )
            )
            Path(template_dir_path, '{{cookiecutter.dir_name}}', 'a_new_file.txt').touch()
            new_file2 = Path(template_dir_path, '{{cookiecutter.dir_name}}', 'new_dir1', 'new_dir2', 'a_file.txt')
            new_file2.parent.mkdir(parents=True)
            new_file2.touch()

            git.add('.', verbose=False)
            git.commit('Template rev 2', verbose=False)
            to_rev = git.get_current_hash(verbose=False)
            to_date = git.get_commit_date(verbose=False)
            assert to_date

            patch_file_path = Path(
                main_temp_path,
                'repo_path',
                '.manageprojects',
                'patches',
                f'{from_rev}_{to_rev}.patch',
            )
            self.assertFalse(patch_file_path.exists())

            # Don't overwrite if destination project git is not clean:
            with RedirectOut() as buffer:
                with self.assertRaises(SystemExit), AssertLogs(self):
                    update_managed_project(
                        project_path=project_path,
                        overwrite=True,  # Update by overwrite
                        password=None,
                        config_file=config_file_path,
                        cleanup=False,  # Keep temp files if this test fails, for better debugging
                        input=False,  # No user input in tests ;)
                    )
            self.assertIn('Abort', buffer.stderr)
            self.assertIn('is not clean', buffer.stderr)
            self.assertIn('?? pyproject.toml', buffer.stdout)

            # Make project git "clean":
            project_git.add('.', verbose=False)
            project_git.commit('add pyproject.toml', verbose=False)

            with RedirectOut() as buffer:
                with AssertLogs(self) as logs:
                    result = update_managed_project(
                        project_path=project_path,
                        overwrite=True,  # Update by overwrite
                        password=None,
                        config_file=config_file_path,
                        cleanup=False,  # Keep temp files if this test fails, for better debugging
                        input=False,  # No user input in tests ;)
                    )
            self.assertEqual(buffer.stderr, '')
            self.assertIn('Update by overwrite', buffer.stdout)
            self.assertIn('UPDATE file', buffer.stdout)
            self.assertIn('NEW file', buffer.stdout)

            self.assert_file_content(
                dst_file_path,
                inspect.cleandoc(
                    '''
                    # This is a test line, not changed
                    #
                    # Revision 2
                    #
                    # The same cookiecutter value:
                    print('Test: FooBar')
                    '''
                ),
            )
            assert_is_file(path=Path(project_path, 'a_new_file.txt'))
            assert_is_file(path=Path(project_path, 'new_dir1', 'new_dir2', 'a_file.txt'))

            self.assertIsInstance(result, OverwriteResult)
            self.assertEqual(
                result,
                OverwriteResult(
                    to_rev=to_rev,
                    to_commit_date=to_date,
                ),
            )

            # Check updated toml file:
            with AssertLogs(self, loggers=('manageprojects',)) as logs:
                toml = PyProjectToml(project_path=project_path)
                mp_meta: ManageProjectsMeta = toml.get_mp_meta()
            logs.assert_in('Read existing pyproject.toml')
            self.assertEqual(mp_meta.initial_revision, from_rev)
            self.assertEqual(mp_meta.applied_migrations, [to_rev])



import logging
from pathlib import Path

from rich import print  # noqa

from manageprojects.git import Git


logger = logging.getLogger(__name__)


def _print_version(module, project_root: Path = None):
    print(f'[bold][green]{module.__name__}[/green] v', end='')
    print(module.__version__, end=' ')

    if not project_root:
        project_root = Path(module.__file__).parent.parent
    git_path = project_root / '.git'
    if git_path.is_dir():
        git = Git(cwd=git_path)
        current_hash = git.get_current_hash(verbose=False)
        print(f'[blue]{current_hash}[/blue]', end=' ')
        print(f'([white]{git.cwd}[/white])')
    else:
        print(f'(No git found for: {project_root})')


def print_version(module, project_root: Path = None) -> None:
    try:
        _print_version(module=module, project_root=project_root)
    except Exception as err:
        logger.exception('Error print version: %s', err)

        # Catch all errors, otherwise the CLI is not usable ;)
        print(f'{module} (print version error: {err})')



import dataclasses
import datetime
from pathlib import Path
from typing import Optional


@dataclasses.dataclass
class CookiecutterResult:
    """
    Store information about a created Cookiecutter template
    """

    destination_path: Path
    git_path: Path
    git_hash: str
    commit_date: Optional[datetime.datetime]
    cookiecutter_context: dict

    def get_comment(self):
        if self.commit_date:
            return self.commit_date.isoformat()
        return ''


@dataclasses.dataclass
class ManageProjectsMeta:
    """
    Information about 'manageprojects' git hashes
    """

    initial_revision: str
    initial_date: datetime.datetime
    applied_migrations: list[str]
    cookiecutter_template: Optional[str]  # CookieCutter Template path or GitHub url
    cookiecutter_directory: Optional[str]  # Directory name of the CookieCutter Template
    cookiecutter_context: Optional[dict]

    def get_last_git_hash(self) -> str:
        """
        Get the lash git hash from 'pyproject.toml' file.
        """
        if self.applied_migrations:
            return self.applied_migrations[-1]
        return self.initial_revision


@dataclasses.dataclass
class ResultBase:
    to_rev: str
    to_commit_date: datetime.datetime


@dataclasses.dataclass
class GenerateTemplatePatchResult(ResultBase):
    repo_path: Path  # Cookiecutter template path
    patch_file_path: Path

    from_rev: str
    compiled_from_path: Path

    compiled_to_path: Path


@dataclasses.dataclass
class OverwriteResult(ResultBase):
    pass



from bx_py_utils.path import assert_is_dir, assert_is_file

from manageprojects.git import Git


def init_git(path, comment='The initial commit ;)', branch_name='main') -> tuple[Git, str]:
    git = Git(cwd=path, detect_root=False)
    git_path = git.init(
        branch_name=branch_name,
        user_name='Mr. Test',
        user_email='foo-bar@test.tld',
        verbose=False,
    )
    assert git_path == path

    assert_is_dir(path / '.git')
    assert_is_file(path / '.git' / 'config')

    assert git.get_config('user.name', verbose=False)
    assert git.get_config('user.email', verbose=False)

    git.add('.', verbose=False)
    git.commit(comment, verbose=False)
    reflog = git.reflog(verbose=False)

    assert ' commit (initial)' in reflog, reflog
    assert comment in reflog, reflog

    git_hash = git.get_current_hash(verbose=False)
    assert f'{git_hash} HEAD@' in reflog, reflog

    # Add a "fake" origin and push the current branch to it.
    # Needed if "darker" is used ;)
    git.git_verbose_check_call('remote', 'add', 'origin', '.')
    git.git_verbose_check_call('push', 'origin', branch_name)

    return git, git_hash



import sys
from pathlib import Path


INITIAL_REVISION = 'initial_revision'
INITIAL_DATE = 'initial_date'
APPLIED_MIGRATIONS = 'applied_migrations'
COOKIECUTTER_TEMPLATE = 'cookiecutter_template'
COOKIECUTTER_DIRECTORY = 'cookiecutter_directory'
COOKIECUTTER_CONTEXT = 'cookiecutter_context'

CLI_EPILOG = 'Project Homepage: https://github.com/jedie/manageprojects'

# Draker has some troubles fixing new lines,
# so just call autopep8 with following selection before call darker:
FORMAT_PY_FILE_DARKER_PRE_FIXES = ','.join(
    sorted(
        [
            'E301',  # expected 1 blank line
            'E302',  # expected 2 blank lines
            'E303',  # too many blank lines
            'W391',  # blank line at end of file
            'E305',  # expected 2 blank lines after class or function definition
        ]
    )
)
FORMAT_PY_FILE_DEFAULT_MIN_PYTON_VERSION = '3.9'
FORMAT_PY_FILE_DEFAULT_MAX_LINE_LENGTH = 119

PY_BIN_PATH = Path(sys.executable).parent



import inspect
import json
from pathlib import Path

import yaml
from bx_py_utils.path import assert_is_dir, assert_is_file
from bx_py_utils.test_utils.datetime import parse_dt

from manageprojects.cli.cli_app import clone_project
from manageprojects.cookiecutter_templates import start_managed_project, update_managed_project
from manageprojects.data_classes import (
    CookiecutterResult,
    GenerateTemplatePatchResult,
    ManageProjectsMeta,
)
from manageprojects.test_utils.click_cli_utils import invoke_click
from manageprojects.test_utils.git_utils import init_git
from manageprojects.test_utils.logs import AssertLogs
from manageprojects.tests.base import BaseTestCase
from manageprojects.utilities.pyproject_toml import PyProjectToml
from manageprojects.utilities.temp_path import TemporaryDirectory


class CookiecutterTemplatesTestCase(BaseTestCase):
    def test_start_managed_project(self):
        repro_name = 'mp_test_template1'
        cookiecutter_template = f'https://github.com/jedie/{repro_name}/'
        directory = 'test_template1'

        with TemporaryDirectory(prefix='manageprojects_') as main_temp_path:
            replay_dir_path = main_temp_path / '.cookiecutter_replay'
            replay_dir_path.mkdir()

            cookiecutters_dir = main_temp_path / '.cookiecutters'
            config = {
                'cookiecutters_dir': str(cookiecutters_dir),
                'replay_dir': str(replay_dir_path),
            }
            config_yaml = yaml.dump(config)
            config_file_path = main_temp_path / 'cookiecutter_config.yaml'
            config_file_path.write_text(config_yaml)

            replay_file_path = replay_dir_path / f'{directory}.json'
            cookiecutter_output_dir = main_temp_path / 'cookiecutter_output_dir'
            assert replay_file_path.exists() is False
            assert cookiecutter_output_dir.exists() is False
            assert cookiecutters_dir.exists() is False

            with AssertLogs(self) as logs:
                result: CookiecutterResult = start_managed_project(
                    template=cookiecutter_template,
                    output_dir=cookiecutter_output_dir,
                    input=False,
                    directory=directory,
                    config_file=config_file_path,
                    extra_context={
                        'dir_name': 'a_dir_name',
                        'file_name': 'a_file_name',
                    },
                )
            logs.assert_in('Cookiecutter generated here', cookiecutter_output_dir)

            self.assertIsInstance(result, CookiecutterResult)
            git_path = cookiecutters_dir / repro_name
            self.assertEqual(result.git_path, git_path)
            assert_is_dir(git_path)  # Checkout was made?
            assert_is_dir(cookiecutter_output_dir)  # Created from cookiecutter ?

            # Cookiecutter creates a replay JSON?
            assert_is_file(replay_file_path)
            replay_json = replay_file_path.read_text()
            replay_data = json.loads(replay_json)
            assert replay_data == {
                'cookiecutter': {
                    'dir_name': 'a_dir_name',
                    'file_name': 'a_file_name',
                    '_template': cookiecutter_template,
                    '_output_dir': str(cookiecutter_output_dir),
                }
            }

            # Our replay config was used?
            assert_is_file(cookiecutter_output_dir / 'a_dir_name' / 'a_file_name.py')

            project_path = cookiecutter_output_dir / 'a_dir_name'

            # pyproject.toml created?
            with AssertLogs(self, loggers=('manageprojects',)) as logs:
                toml = PyProjectToml(project_path=project_path)
                result: ManageProjectsMeta = toml.get_mp_meta()
            logs.assert_in('Read existing pyproject.toml')
            self.assertIsInstance(result, ManageProjectsMeta)
            self.assertEqual(
                result,
                ManageProjectsMeta(
                    initial_revision='42f18f3',
                    initial_date=parse_dt('2022-11-02T19:40:09+01:00'),
                    applied_migrations=[],
                    cookiecutter_template='https://github.com/jedie/mp_test_template1/',
                    cookiecutter_directory='test_template1',
                    cookiecutter_context={
                        'cookiecutter': {
                            'dir_name': 'a_dir_name',
                            'file_name': 'a_file_name',
                            '_template': 'https://github.com/jedie/mp_test_template1/',
                        }
                    },
                ),
            )

            ###################################################################################
            # Test clone a existing project

            cloned_path = main_temp_path / 'cloned_project'
            with AssertLogs(self) as logs:
                output = invoke_click(clone_project, project_path, cloned_path)
                logs.assert_in(
                    'Read existing pyproject.toml',
                    "Call 'cookiecutter'",
                    'Create new pyproject.toml',
                )
            end_path = cloned_path / 'a_dir_name'
            self.assert_in_content(got=output, parts=(f'{project_path} successfully cloned to {end_path}',))

            # pyproject.toml created?
            with AssertLogs(self, loggers=('manageprojects',)) as logs:
                toml = PyProjectToml(project_path=end_path)
                result: ManageProjectsMeta = toml.get_mp_meta()
            logs.assert_in('Read existing pyproject.toml')
            self.assertIsInstance(result, ManageProjectsMeta)
            self.assertEqual(
                result,
                ManageProjectsMeta(
                    initial_revision='42f18f3',
                    initial_date=parse_dt('2022-11-02T19:40:09+01:00'),
                    applied_migrations=[],
                    cookiecutter_template='https://github.com/jedie/mp_test_template1/',
                    cookiecutter_directory='test_template1',
                    cookiecutter_context={
                        'cookiecutter': {
                            'dir_name': 'a_dir_name',
                            'file_name': 'a_file_name',
                            '_template': 'https://github.com/jedie/mp_test_template1/',
                        }
                    },
                ),
            )

    def test_update_project(self):
        cookiecutter_context = {
            'dir_name': 'a_directory',
            'file_name': 'a_file_name',
            'value': 'FooBar',
        }
        context = {'cookiecutter': cookiecutter_context}

        with TemporaryDirectory(prefix='test_generate_template_patch_') as main_temp_path:
            project_path = main_temp_path / 'project_path'
            dst_file_path = project_path / 'a_file_name.py'

            template_path = main_temp_path / 'template'
            template_dir_name = 'template_dir'
            template_dir_path = template_path / template_dir_name
            config_file_path = template_dir_path / 'cookiecutter.json'

            test_file_path = (
                template_dir_path / '{{cookiecutter.dir_name}}' / '{{cookiecutter.file_name}}.py'
            )

            dst_file_path.parent.mkdir(parents=True)
            dst_file_path.write_text('# This is a test line, not changed')

            config_file_path.parent.mkdir(parents=True)
            config_file_path.write_text(json.dumps(cookiecutter_context))

            project_git, project_from_rev = init_git(project_path, comment='Git init project.')
            dst_file_path.write_text(
                inspect.cleandoc(
                    '''
                    # This is a test line, not changed
                    #
                    # Revision 1
                    #
                    # The same cookiecutter value:
                    print('Test: FooBar')
                    '''
                )
            )
            project_git.add('.')
            project_git.commit('Store revision 1')

            test_file_path.parent.mkdir(parents=True)
            test_file_path.write_text(
                inspect.cleandoc(
                    '''
                    # This is a test line, not changed
                    #
                    # Revision 1
                    #
                    # The same cookiecutter value:
                    print('Test: {{ cookiecutter.value }}')
                    '''
                )
            )

            git, from_rev = init_git(template_path, comment='Git init template.')
            from_date = git.get_commit_date(verbose=False)

            with AssertLogs(self, loggers=('manageprojects',)) as logs:
                toml = PyProjectToml(project_path=project_path)
                toml.init(
                    revision=from_rev,
                    dt=from_date,
                    template=str(template_path),
                    directory=template_dir_name,
                )
            logs.assert_in('Create new pyproject.toml')
            toml.create_or_update_cookiecutter_context(context=context)
            toml.save()
            toml_content = toml.path.read_text()
            self.assertIn('# Created by manageprojects', toml_content)
            self.assertIn(
                '[manageprojects] # https://github.com/jedie/manageprojects', toml_content
            )
            self.assertIn(f'initial_revision = "{from_rev}"', toml_content)
            self.assertIn(f'cookiecutter_template = "{template_path}"', toml_content)
            self.assertIn('cookiecutter_directory = "template_dir"', toml_content)

            test_file_path.write_text(
                inspect.cleandoc(
                    '''
                    # This is a test line, not changed
                    #
                    # Revision 2
                    #
                    # The same cookiecutter value:
                    print('Test: {{ cookiecutter.value }}')
                    '''
                )
            )
            git.add('.', verbose=False)
            git.commit('Template rev 2', verbose=False)
            to_rev = git.get_current_hash(verbose=False)
            to_date = git.get_commit_date(verbose=False)
            assert to_date

            patch_file_path = Path(
                main_temp_path,
                'repo_path',
                '.manageprojects',
                'patches',
                f'{from_rev}_{to_rev}.patch',
            )
            self.assertFalse(patch_file_path.exists())

            with AssertLogs(self) as logs:
                result = update_managed_project(
                    project_path=project_path,
                    password=None,
                    config_file=config_file_path,
                    cleanup=False,  # Keep temp files if this test fails, for better debugging
                    input=False,  # No user input in tests ;)
                )
            logs.assert_in(
                'No temp files cleanup',
                'Cookiecutter generated',
                'Read existing pyproject.toml',
            )

            self.assert_file_content(
                dst_file_path,
                inspect.cleandoc(
                    '''
                    # This is a test line, not changed
                    #
                    # Revision 2
                    #
                    # The same cookiecutter value:
                    print('Test: FooBar')
                    '''
                ),
            )

            self.assertIsInstance(result, GenerateTemplatePatchResult)
            patch_file_path = Path(
                project_path,
                '.manageprojects',
                'patches',
                f'{result.from_rev}_{result.to_rev}.patch',
            )
            assert_is_file(patch_file_path)
            patch_temp_path = result.compiled_to_path.parent
            self.assertEqual(
                result,
                GenerateTemplatePatchResult(
                    repo_path=template_dir_path,
                    patch_file_path=patch_file_path,
                    from_rev=from_rev,
                    compiled_from_path=patch_temp_path / f'{result.from_rev}_compiled',
                    to_rev=to_rev,
                    to_commit_date=to_date,
                    compiled_to_path=patch_temp_path / 'to_rev_compiled',
                ),
            )

            # Check updated toml file:
            with AssertLogs(self, loggers=('manageprojects',)) as logs:
                toml = PyProjectToml(project_path=project_path)
                mp_meta: ManageProjectsMeta = toml.get_mp_meta()
            logs.assert_in('Read existing pyproject.toml')
            self.assertEqual(mp_meta.initial_revision, from_rev)
            self.assertEqual(mp_meta.applied_migrations, [to_rev])



import datetime
import shutil
from collections.abc import Iterable
from pathlib import Path
from pprint import pprint
from unittest import TestCase

from bx_py_utils.path import assert_is_file

import manageprojects
from manageprojects.utilities.pyproject_toml import TomlDocument, get_toml_document


PROJECT_PATH = Path(manageprojects.__file__).parent.parent
GIT_BIN_PARENT = Path(shutil.which('git')).parent


class BaseTestCase(TestCase):
    maxDiff = None

    def assert_toml(self, path: Path, expected: dict):
        toml_document: TomlDocument = get_toml_document(path)
        got = toml_document.doc.unwrap()  # TOMLDocument -> dict
        try:
            self.assertDictEqual(got, expected)
        except AssertionError:
            print('-' * 79)
            pprint(got)
            print('-' * 79)
            raise

    def assert_datetime_range(
        self, dt: datetime.datetime, reference: datetime.datetime, max_diff_sec: int = 5
    ):
        self.assertEqual(dt.tzinfo, reference.tzinfo)
        diff = datetime.timedelta(seconds=max_diff_sec)
        min_dt = reference - diff
        max_dt = reference + diff
        self.assertLessEqual(dt, max_dt)
        self.assertGreaterEqual(dt, min_dt)

    def assert_datetime_now_range(self, dt: datetime.datetime, max_diff_sec: int = 5):
        now = datetime.datetime.now(tz=dt.tzinfo)
        self.assert_datetime_range(dt=dt, reference=now, max_diff_sec=max_diff_sec)

    def assert_content(self, got: str, expected: str, strip=True):
        if strip:
            got = got.strip()
            expected = expected.strip()

        try:
            self.assertEqual(got, expected)
        except AssertionError:
            print('-' * 79)
            print(got)
            print('-' * 79)
            raise

    def assert_in_content(self, *, got: str, parts: Iterable[str]):
        assert parts
        missing_parts = [part for part in parts if part not in got]
        if missing_parts:
            print('-' * 79)
            print(got)
            print('-' * 79)
            info = ', '.join(repr(part) for part in missing_parts)
            raise AssertionError(f'Text parts: {info} not found in: {got!r}')

    def assert_file_content(self, path: Path, content: str):
        assert_is_file(path)
        file_content = path.read_text().strip()
        self.assert_content(got=file_content, expected=content)



import dataclasses
import subprocess
from typing import Callable, Optional, Union
from unittest.mock import patch

from bx_py_utils.test_utils.context_managers import MassContextManager


@dataclasses.dataclass
class Call:
    popenargs: Union[tuple, list]
    args: tuple
    kwargs: Optional[dict]


class FakeStdout:
    def __init__(self, *, stdout: str):
        self.stdout = stdout


class SimpleRunReturnCallback:
    """
    Helper to return the same output on every mocked subprocess.run() call.
    """

    def __init__(self, *, stdout: str):
        self.stdout = stdout

    def __call__(self, *args, **kwargs):
        return FakeStdout(stdout=self.stdout)


class MissingSubprocessCallMockCallback:
    def __call__(self, popenargs, *args, **kwargs):
        raise AttributeError(f'SubprocessCallMock needs a "return_callback" for "subprocess.run({popenargs})" usage!')


class SubprocessCallMock(MassContextManager):
    def __init__(self, return_callback: Optional[Callable] = None):
        if return_callback is None:
            return_callback = MissingSubprocessCallMockCallback()

        class CallHelper:
            def __init__(self, calls, return_callback=None):
                self.calls = calls
                self.return_callback = return_callback

            def __call__(self, popenargs, *args, **kwargs):
                self.calls.append(Call(popenargs, args, kwargs))
                if self.return_callback:
                    return self.return_callback(popenargs, args, kwargs)

        self.calls: list[Call] = []
        self.mocks = [
            patch.object(subprocess, 'call', CallHelper(self.calls)),
            patch.object(subprocess, 'run', CallHelper(self.calls, return_callback)),
        ]

    def get_popenargs(self, rstrip_paths: Optional[tuple] = None) -> list:
        if rstrip_paths:
            rstrip_paths = [str(item) for item in rstrip_paths if item]  # e.g.: Path -> str

        result = []
        for call in self.calls:
            if rstrip_paths:
                temp = []
                for arg in call.popenargs:
                    for rstrip_path in rstrip_paths:
                        if arg.startswith(rstrip_path):
                            arg = f'...{arg.removeprefix(rstrip_path)}'
                    temp.append(arg)

                result.append(temp)

            else:
                result.append(call.popenargs)

        return result



import inspect
from pathlib import Path

from bx_py_utils.path import assert_is_dir

from manageprojects.cookiecutter_api import get_repo_path
from manageprojects.test_utils.logs import AssertLogs
from manageprojects.tests.base import BaseTestCase


class CookiecutterApiTestCase(BaseTestCase):
    def test_get_repo_path(self):
        repro_name = 'mp_test_template1'
        cookiecutter_template = f'https://github.com/jedie/{repro_name}/'
        directory = 'test_template1'

        with AssertLogs(self) as logs:
            repo_path = get_repo_path(
                template=cookiecutter_template,
                directory=directory,
                checkout='84d23bf',  # Old version
            )
        logs.assert_in('repo_dir', repo_path)

        self.assertIsInstance(repo_path, Path)
        self.assertEqual(repo_path.name, directory)
        assert_is_dir(repo_path.parent / '.git')

        test_file_path = Path(
            repo_path, '{{cookiecutter.dir_name}}', '{{cookiecutter.file_name}}.py'
        )
        self.assert_file_content(
            test_file_path,
            inspect.cleandoc(
                '''
                print('Hello World 1')
                '''
            ),
        )

        with AssertLogs(self) as logs:
            repo_path = get_repo_path(
                template=cookiecutter_template,
                directory=directory,
                checkout=None,  # Current main branch
            )
        logs.assert_in('repo_dir', repo_path)

        self.assertIsInstance(repo_path, Path)
        self.assertEqual(repo_path.name, directory)
        assert_is_dir(repo_path.parent / '.git')

        test_file_path = Path(
            repo_path, '{{cookiecutter.dir_name}}', '{{cookiecutter.file_name}}.py'
        )
        self.assert_file_content(
            test_file_path,
            inspect.cleandoc(
                '''
                # A later added Comment...

                print('Hello World 1')
                '''
            ),
        )



import logging
import shutil
import tempfile
from pathlib import Path


logger = logging.getLogger(__name__)


class TemporaryDirectory:
    """
    Similar to origin tempfile.TemporaryDirectory, but:
     * cleanup only if no error happens
     * returns a Path object
    """

    def __init__(self, prefix=None, suffix=None, dir=None, cleanup=True):
        self.prefix = prefix
        self.suffix = suffix
        self.dir = dir
        self.cleanup = cleanup

    def __enter__(self) -> Path:
        temp_name = tempfile.mkdtemp(prefix=self.prefix, suffix=self.suffix, dir=self.dir)
        self.temp_path = Path(temp_name)
        return self.temp_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            logger.exception(f"no cleanup of {self.temp_path}, cause: {exc_val}")
            return False

        if self.cleanup:
            shutil.rmtree(self.temp_path)
        else:
            logger.warning('No temp files cleanup for: %s', self.temp_path)



import subprocess

from bx_py_utils.path import assert_is_file
from packaging.version import Version

from manageprojects import __version__
from manageprojects.cli.cli_app import PACKAGE_ROOT
from manageprojects.test_utils.click_cli_utils import subprocess_cli
from manageprojects.test_utils.project_setup import check_editor_config, get_py_max_line_length
from manageprojects.tests.base import BaseTestCase
from manageprojects.utilities import code_style


class ProjectSetupTestCase(BaseTestCase):
    def test_version(self):
        self.assertIsNotNone(__version__)

        version = Version(__version__)  # Will raise InvalidVersion() if wrong formatted
        self.assertEqual(str(version), __version__)

        cli_bin = PACKAGE_ROOT / 'cli.py'
        assert_is_file(cli_bin)

        output = subprocess.check_output([cli_bin, 'version'], text=True)
        self.assertIn(f'manageprojects v{__version__}', output)

    def test_code_style(self):
        cli_bin = PACKAGE_ROOT / 'cli.py'
        assert_is_file(cli_bin)

        try:
            output = subprocess_cli(
                cli_bin=cli_bin,
                args=('check-code-style',),
                exit_on_error=False,
            )
        except subprocess.CalledProcessError as err:
            self.assertIn('.venv/bin/darker', err.stdout)  # darker was called?
        else:
            if 'Code style: OK' in output:
                self.assertIn('.venv/bin/darker', output)  # darker was called?
                return  # Nothing to fix -> OK

        # Try to "auto" fix code style:

        try:
            output = subprocess_cli(
                cli_bin=cli_bin,
                args=('fix-code-style',),
                exit_on_error=False,
            )
        except subprocess.CalledProcessError as err:
            output = err.stdout

        self.assertIn('.venv/bin/darker', output)  # darker was called?

        # Check again and display the output:

        try:
            code_style.check(package_root=PACKAGE_ROOT)
        except SystemExit as err:
            self.assertEqual(err.code, 0, 'Code style error, see output above!')

    def test_check_editor_config(self):
        check_editor_config(package_root=PACKAGE_ROOT)

        max_line_length = get_py_max_line_length(package_root=PACKAGE_ROOT)
        self.assertEqual(max_line_length, 119)



import logging
import os
from pathlib import Path


logger = logging.getLogger(__name__)


def get_user_config_path() -> Path:
    if config_dir := os.environ.get('XDG_CONFIG_HOME'):
        config_path = Path(config_dir)
        if config_path.is_dir():
            return config_path

    config_path = Path.home() / '.config'
    if config_path.is_dir():
        return config_path

    config_path = Path.home() / 'Library' / 'Preferences'
    if config_path.is_dir():
        return config_path

    config_path = Path.home()
    logger.warning('Fallback user config path to: %s', config_path)
    return config_path


def get_mp_config_path() -> Path:
    user_config_path = get_user_config_path()
    mp_config_path = user_config_path / 'manageprojects'
    mp_config_path.mkdir(exist_ok=True)
    return mp_config_path



import tempfile
from pathlib import Path


class TempContentFile:
    def __init__(self, content, prefix=None, suffix=None):
        self.content = content
        self.prefix = prefix
        self.suffix = suffix

    def __enter__(self) -> Path:
        self.tmp = tempfile.NamedTemporaryFile(prefix=self.prefix, suffix=self.suffix)
        self.tmp.__enter__()
        file_path = Path(self.tmp.name)
        file_path.write_text(self.content)
        return file_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tmp.__exit__(exc_type, exc_val, exc_tb)



import logging
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Optional

import rich_click as click
from bx_py_utils.path import assert_is_dir, assert_is_file
from rich import print  # noqa
from rich_click import RichGroup

import manageprojects
from manageprojects import constants
from manageprojects.constants import (
    FORMAT_PY_FILE_DARKER_PRE_FIXES,
    FORMAT_PY_FILE_DEFAULT_MAX_LINE_LENGTH,
    FORMAT_PY_FILE_DEFAULT_MIN_PYTON_VERSION,
)
from manageprojects.cookiecutter_templates import (
    clone_managed_project,
    reverse_managed_project,
    start_managed_project,
    update_managed_project,
)
from manageprojects.data_classes import CookiecutterResult
from manageprojects.format_file import format_one_file
from manageprojects.utilities import code_style
from manageprojects.utilities.log_utils import log_config
from manageprojects.utilities.publish import publish_package
from manageprojects.utilities.subprocess_utils import verbose_check_call
from manageprojects.utilities.version_info import print_version


logger = logging.getLogger(__name__)


PACKAGE_ROOT = Path(manageprojects.__file__).parent.parent
assert_is_file(PACKAGE_ROOT / 'pyproject.toml')

OPTION_ARGS_DEFAULT_TRUE = dict(is_flag=True, show_default=True, default=True)
OPTION_ARGS_DEFAULT_FALSE = dict(is_flag=True, show_default=True, default=False)
ARGUMENT_EXISTING_DIR = dict(
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, path_type=Path)
)
ARGUMENT_NOT_EXISTING_DIR = dict(
    type=click.Path(exists=False, file_okay=False, dir_okay=True, readable=False, writable=True, path_type=Path)
)
ARGUMENT_EXISTING_FILE = dict(
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=Path)
)


class ClickGroup(RichGroup):  # FIXME: How to set the "info_name" easier?
    def make_context(self, info_name, *args, **kwargs):
        info_name = './cli.py'
        return super().make_context(info_name, *args, **kwargs)


@click.group(
    cls=ClickGroup,
    epilog=constants.CLI_EPILOG,
)
def cli():
    pass


@click.command()
@click.option('--verbose/--no-verbose', **OPTION_ARGS_DEFAULT_FALSE)
def mypy(verbose: bool = True):
    """Run Mypy (configured in pyproject.toml)"""
    verbose_check_call('mypy', '.', cwd=PACKAGE_ROOT, verbose=verbose, exit_on_error=True)


cli.add_command(mypy)


@click.command()
@click.option('--verbose/--no-verbose', **OPTION_ARGS_DEFAULT_FALSE)
def coverage(verbose: bool = True):
    """
    Run and show coverage.
    """
    verbose_check_call('coverage', 'run', verbose=verbose, exit_on_error=True)
    verbose_check_call('coverage', 'combine', '--append', verbose=verbose, exit_on_error=True)
    verbose_check_call('coverage', 'report', '--fail-under=50', verbose=verbose, exit_on_error=True)
    verbose_check_call('coverage', 'xml', verbose=verbose, exit_on_error=True)
    verbose_check_call('coverage', 'json', verbose=verbose, exit_on_error=True)


cli.add_command(coverage)


@click.command()
def install():
    """
    Run pip-sync and install 'manageprojects' via pip as editable.
    """
    verbose_check_call('pip-sync', PACKAGE_ROOT / 'requirements.dev.txt')
    verbose_check_call('pip', 'install', '--no-deps', '-e', '.')


cli.add_command(install)


@click.command()
def safety():
    """
    Run safety check against current requirements files
    """
    verbose_check_call('safety', 'check', '-r', 'requirements.dev.txt')


cli.add_command(safety)


@click.command()
def update():
    """
    Update "requirements*.txt" dependencies files
    """
    bin_path = Path(sys.executable).parent

    verbose_check_call(bin_path / 'pip', 'install', '-U', 'pip')
    verbose_check_call(bin_path / 'pip', 'install', '-U', 'pip-tools')

    extra_env = dict(
        CUSTOM_COMPILE_COMMAND='./cli.py update',
    )

    pip_compile_base = [
        bin_path / 'pip-compile',
        '--verbose',
        '--allow-unsafe',  # https://pip-tools.readthedocs.io/en/latest/#deprecations
        '--resolver=backtracking',  # https://pip-tools.readthedocs.io/en/latest/#deprecations
        '--upgrade',
        '--generate-hashes',
    ]

    # Only "prod" dependencies:
    verbose_check_call(
        *pip_compile_base,
        'pyproject.toml',
        '--output-file',
        'requirements.txt',
        extra_env=extra_env,
    )

    # dependencies + "dev"-optional-dependencies:
    verbose_check_call(
        *pip_compile_base,
        'pyproject.toml',
        '--extra=dev',
        '--output-file',
        'requirements.dev.txt',
        extra_env=extra_env,
    )

    verbose_check_call(bin_path / 'safety', 'check', '-r', 'requirements.dev.txt')

    # Install new dependencies in current .venv:
    verbose_check_call(bin_path / 'pip-sync', 'requirements.dev.txt')


cli.add_command(update)


@click.command()
@click.argument('template')
@click.argument('output_dir', **ARGUMENT_NOT_EXISTING_DIR)
@click.option(
    '--directory',
    default=None,
    help=(
        'Cookiecutter Option: Directory within repo that holds cookiecutter.json file'
        ' for advanced repositories with multi templates in it'
    ),
)
@click.option(
    '--checkout',
    default=None,
    help='Cookiecutter Option: branch, tag or commit to checkout after git clone',
)
@click.option(
    '--input/--no-input',
    **OPTION_ARGS_DEFAULT_TRUE,
    help=('Cookiecutter Option: Do not prompt for parameters' ' and only use cookiecutter.json file content'),
)
@click.option(
    '--replay/--no-replay',
    **OPTION_ARGS_DEFAULT_FALSE,
    help=('Cookiecutter Option: Do not prompt for parameters' ' and only use information entered previously'),
)
@click.option(
    '--password',
    default=None,
    help='Cookiecutter Option: Password to use when extracting the repository',
)
@click.option(
    '--config-file',
    default=None,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    help='Cookiecutter Option: Optional path to "cookiecutter_config.yaml"',
)
def start_project(
    template: str,
    output_dir: Path,
    directory: Optional[str],
    checkout: Optional[str],
    input: bool,
    replay: bool,
    password: Optional[str],
    config_file: Optional[Path],
):
    """
    Start a new "managed" project via a CookieCutter Template.
    Note: The CookieCutter Template *must* be use git!

    e.g.:

    ./cli.py start-project https://github.com/jedie/cookiecutter_templates/ --directory piptools-python ~/foobar/
    """
    log_config()
    print(f'Start project with template: {template!r}')
    print(f'Destination: {output_dir}')
    if output_dir.exists():
        print(f'Error: Destination "{output_dir}" already exists')
        sys.exit(1)
    if not output_dir.parent.is_dir():
        print(f'Error: Destination parent "{output_dir.parent}" does not exists')
        sys.exit(1)

    result: CookiecutterResult = start_managed_project(
        template=template,
        checkout=checkout,
        output_dir=output_dir,
        input=input,
        replay=replay,
        password=password,
        directory=directory,
        config_file=config_file,
    )

    print(
        f'CookieCutter template {template!r}'
        f' with git hash {result.git_hash}'
        f' was created here: {output_dir}'
    )
    return result


cli.add_command(start_project)


@click.command()
@click.argument('project_path', **ARGUMENT_EXISTING_DIR)
@click.option(
    '--overwrite/--no-overwrite',
    **OPTION_ARGS_DEFAULT_TRUE,
    help=(
        'Overwrite all Cookiecutter template files to the last template state and'
        ' do not apply the changes via git patches.'
        ' The developer is supposed to apply the differences manually via git.'
        ' Will be aborted if the project git repro is not in a clean state.'
    ),
)
@click.option(
    '--password',
    default=None,
    help='Cookiecutter Option: Password to use when extracting the repository',
)
@click.option(
    '--config-file',
    default=None,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    help='Cookiecutter Option: Optional path to "cookiecutter_config.yaml"',
)
@click.option(
    '--input/--no-input',
    **OPTION_ARGS_DEFAULT_FALSE,
    help='Cookiecutter Option: Do not prompt for parameters and only use cookiecutter.json file content',
)
@click.option(
    '--cleanup/--no-cleanup',
    **OPTION_ARGS_DEFAULT_TRUE,
    help='Cleanup created temporary files',
)
def update_project(
    project_path: Path,
    overwrite: bool,
    password: Optional[str],
    config_file: Optional[Path],
    input: bool,
    cleanup: bool,
):
    """
    Update a existing project.

    e.g. update by overwrite (and merge changes manually via git):

    ./cli.py update-project ~/foo/bar/
    """
    log_config()
    print(f'Update project: "{project_path}"...')
    update_managed_project(
        project_path=project_path,
        overwrite=overwrite,
        password=password,
        config_file=config_file,
        cleanup=cleanup,
        input=input,
    )
    print(f'Managed project "{project_path}" updated, ok.')


cli.add_command(update_project)


@click.command()
@click.argument('project_path', **ARGUMENT_EXISTING_DIR)
@click.argument('output_dir', **ARGUMENT_NOT_EXISTING_DIR)
@click.option(
    '--password',
    default=None,
    help='Cookiecutter Option: Password to use when extracting the repository',
)
@click.option(
    '--config-file',
    default=None,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    help='Cookiecutter Option: Optional path to "cookiecutter_config.yaml"',
)
@click.option(
    '--input/--no-input',
    **OPTION_ARGS_DEFAULT_FALSE,
    help=('Cookiecutter Option: Do not prompt for parameters' ' and only use cookiecutter.json file content'),
)
def clone_project(
    project_path: Path,
    output_dir: Path,
    input: bool,
    checkout: Optional[str] = None,
    password: Optional[str] = None,
    config_file: Optional[Path] = None,
):
    """
    Clone existing project by replay the cookiecutter template in a new directory.

    e.g.:

    ./cli.py clone-project ~/foo/bar ~/cloned/
    """
    print(locals())
    log_config()
    return clone_managed_project(
        project_path=project_path,
        destination=output_dir,
        checkout=checkout,
        password=password,
        config_file=config_file,
        input=input,
    )


cli.add_command(clone_project)


@click.command()
@click.argument('project_path', **ARGUMENT_EXISTING_DIR)
@click.argument('destination', **ARGUMENT_NOT_EXISTING_DIR)
@click.option(
    '--overwrite/--no-overwrite',
    **OPTION_ARGS_DEFAULT_FALSE,
    help='Overwrite existing files.',
)
def reverse(
    project_path: Path,
    destination: Path,
    overwrite: bool,
):
    """
    Create a cookiecutter template from a managed project.

    e.g.:

    ./cli.py reverse ~/my_managed_project/ ~/my_new_cookiecutter_template/
    """
    log_config()
    return reverse_managed_project(
        project_path=project_path,
        destination=destination,
        overwrite=overwrite,
    )


cli.add_command(reverse)


@click.command()
@click.argument('project_path', **ARGUMENT_EXISTING_DIR)
@click.option(
    '--words/--no-words',
    **OPTION_ARGS_DEFAULT_FALSE,
    help='wiggle Option: word-wise diff and merge.',
)
def wiggle(project_path: Path, words: bool):
    """
    Run wiggle to merge *.rej in given directory.
    https://github.com/neilbrown/wiggle

    e.g.:

    ./cli.py wiggle ~/my_managed_project/
    """
    wiggle_bin = shutil.which('wiggle')
    if not wiggle_bin:
        print('Error: "wiggle" can not be found!')
        print('Hint: sudo apt get install wiggle')
        sys.exit(1)

    assert_is_dir(project_path)

    args = [wiggle_bin, '--merge']
    if words:
        args.append('--words')
    args.append('--replace')

    for rej_file_path in project_path.rglob('*.rej'):
        real_file_path = rej_file_path.with_suffix('')
        if not real_file_path.is_file():
            print(f'Error real file "{real_file_path}" from "{rej_file_path}" not found. Skip.')
            continue
        try:
            verbose_check_call(
                *args,
                real_file_path,
                rej_file_path,
                verbose=True,
                exit_on_error=False,
            )
        except subprocess.CalledProcessError:
            continue


cli.add_command(wiggle)


@click.command()
def publish():
    """
    Build and upload this project to PyPi
    """
    verbose_check_call('pip', 'install', '-e', '.')  # Auto fix: Maybe the current version is not installed ;)

    _run_unittest_cli(verbose=False, exit_after_run=False)  # Don't publish a broken state

    publish_package(
        module=manageprojects,
        package_path=PACKAGE_ROOT,
    )


cli.add_command(publish)


@click.command()
@click.option('--color/--no-color', **OPTION_ARGS_DEFAULT_TRUE)
@click.option('--verbose/--no-verbose', **OPTION_ARGS_DEFAULT_FALSE)
def fix_code_style(color: bool = True, verbose: bool = False):
    """
    Fix code style of all manageprojects source code files via darker
    """
    code_style.fix(package_root=PACKAGE_ROOT, color=color, verbose=verbose)


cli.add_command(fix_code_style)


@click.command()
@click.option('--color/--no-color', **OPTION_ARGS_DEFAULT_TRUE)
@click.option('--verbose/--no-verbose', **OPTION_ARGS_DEFAULT_FALSE)
def check_code_style(color: bool = True, verbose: bool = False):
    """
    Check code style by calling darker + flake8
    """
    code_style.check(package_root=PACKAGE_ROOT, color=color, verbose=verbose)


cli.add_command(check_code_style)


@click.command()
def update_test_snapshot_files():
    """
    Update all test snapshot files (by remove and recreate all snapshot files)
    """

    def iter_snapshot_files():
        yield from PACKAGE_ROOT.rglob('*.snapshot.*')

    removed_file_count = 0
    for item in iter_snapshot_files():
        item.unlink()
        removed_file_count += 1
    print(f'{removed_file_count} test snapshot files removed... run tests...')

    # Just recreate them by running tests:
    _run_unittest_cli(
        extra_env=dict(
            RAISE_SNAPSHOT_ERRORS='0',  # Recreate snapshot files without error
        ),
        verbose=False,
        exit_after_run=False,
    )

    new_files = len(list(iter_snapshot_files()))
    print(f'{new_files} test snapshot files created, ok.\n')


cli.add_command(update_test_snapshot_files)


def _run_unittest_cli(extra_env=None, verbose=True, exit_after_run=True):
    """
    Call the origin unittest CLI and pass all args to it.
    """
    if extra_env is None:
        extra_env = dict()

    extra_env.update(
        dict(
            PYTHONUNBUFFERED='1',
            PYTHONWARNINGS='always',
        )
    )

    args = sys.argv[2:]
    if not args:
        if verbose:
            args = ('--verbose', '--locals', '--buffer')
        else:
            args = ('--locals', '--buffer')

    verbose_check_call(
        sys.executable,
        '-m',
        'unittest',
        *args,
        timeout=15 * 60,
        extra_env=extra_env,
    )
    if exit_after_run:
        sys.exit(0)


cli.add_command(update_test_snapshot_files)


@click.command()  # Dummy command
def test():
    """
    Run unittests
    """
    _run_unittest_cli()


cli.add_command(test)


def _run_tox():
    verbose_check_call(sys.executable, '-m', 'tox', *sys.argv[2:])
    sys.exit(0)


@click.command()  # Dummy "tox" command
def tox():
    """
    Run tox
    """
    _run_tox()


cli.add_command(tox)


@click.command()
@click.option(
    '--py-version',
    default=FORMAT_PY_FILE_DEFAULT_MIN_PYTON_VERSION,
    show_default=True,
    help='Fallback Python version for darker/pyupgrade, if version is not defined in pyproject.toml',
)
@click.option(
    '-l',
    '--max-line-length',
    default=FORMAT_PY_FILE_DEFAULT_MAX_LINE_LENGTH,
    type=int,
    show_default=True,
    help='Fallback max. line length for darker/isort etc., if not defined in .editorconfig',
)
@click.option(
    '--darker-prefixes',
    default=FORMAT_PY_FILE_DARKER_PRE_FIXES,
    show_default=True,
    help='Apply prefixes via autopep8 before calling darker.',
)
@click.argument('file_path', **ARGUMENT_EXISTING_FILE)
def format_file(
    *,
    py_version: str,
    max_line_length: int,
    darker_prefixes: str,
    file_path: Path,
):
    """
    Format and check the given python source code file with darker/autoflake/isort/pyupgrade/autopep8/mypy etc.

    The optional fallback values will be only used, if we can't get them from the project meta files
    like ".editorconfig" and "pyproject.toml"
    """
    format_one_file(
        default_min_py_version=py_version,
        default_max_line_length=max_line_length,
        darker_prefixes=darker_prefixes,
        file_path=file_path,
    )


cli.add_command(format_file)


@click.command()
def version():
    """Print version and exit"""
    # Pseudo command, because the version always printed on every CLI call ;)
    sys.exit(0)


cli.add_command(version)


def main():
    print_version(manageprojects)

    if len(sys.argv) >= 2:
        # Check if we just pass a command call
        command = sys.argv[1]
        if command == 'test':
            _run_unittest_cli()
        elif command == 'tox':
            _run_tox()

    # Execute Click CLI:
    cli.name = './cli.py'
    cli()



import inspect
import sys
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from packaging.version import Version

from manageprojects.test_utils.git_utils import init_git
from manageprojects.test_utils.logs import AssertLogs
from manageprojects.test_utils.subprocess import FakeStdout, SubprocessCallMock
from manageprojects.tests.base import GIT_BIN_PARENT
from manageprojects.utilities import subprocess_utils
from manageprojects.utilities.publish import PublisherGit, build, get_pyproject_toml_version
from manageprojects.utilities.temp_path import TemporaryDirectory


class PublishTestCase(TestCase):
    def test_build(self):
        def return_callback(popenargs, args, kwargs):
            return FakeStdout(stdout='Mocked run output')

        with TemporaryDirectory(prefix='test_build') as temp_path, SubprocessCallMock(return_callback) as call_mock:
            build(package_path=temp_path)

        self.assertEqual(
            call_mock.get_popenargs(rstrip_paths=(str(Path(sys.executable).parent),)),
            [
                ['.../python', '-m', 'build'],
            ],
        )

        # Use poetry:

        def prepare_popenargs_mock(popenargs, cwd=None):
            return popenargs, cwd

        def make_relative_path_mock(path, **kwargs):
            return path

        with TemporaryDirectory(prefix='test_build') as temp_path, SubprocessCallMock(
            return_callback
        ) as call_mock, patch.object(subprocess_utils, 'prepare_popenargs', prepare_popenargs_mock), patch.object(
            subprocess_utils, 'make_relative_path', make_relative_path_mock
        ):
            Path(temp_path / 'poetry.lock').touch()

            build(package_path=temp_path)

        self.assertEqual(
            call_mock.get_popenargs(),
            [
                ['poetry', 'build'],
            ],
        )

    def test_get_pyproject_toml_version(self):
        with TemporaryDirectory(prefix='test_get_pyproject_toml_version') as temp_path:
            pyproject_toml_path = temp_path / 'pyproject.toml'
            pyproject_toml_path.touch()

            version = get_pyproject_toml_version(temp_path)
            self.assertIs(version, None)

            pyproject_toml_path.write_text(
                inspect.cleandoc(
                    '''
                    [project]
                    name = "foo"
                    version = "1.2"
                    description = "bar"
                    '''
                )
            )
            version = get_pyproject_toml_version(temp_path)
            self.assertEqual(version, Version('1.2'))

            pyproject_toml_path.write_text(
                inspect.cleandoc(
                    '''
                    [tool.poetry]
                    name = "foo"
                    version = "2.3"
                    description = "bar"
                    '''
                )
            )
            version = get_pyproject_toml_version(temp_path)
            self.assertEqual(version, Version('2.3'))

    def test_publisher_fast_checks_not_existing_tag(self):
        # https://github.com/jedie/manageprojects/issues/68
        with AssertLogs(self, loggers=('manageprojects',)), TemporaryDirectory(
            prefix='test_publisher_fast_checks_not_existing_tag'
        ) as temp_path:
            Path(temp_path, '1.txt').touch()
            init_git(temp_path, comment='The initial commit ;)')

            pgit = PublisherGit(
                package_path=temp_path,
                version=Version('0.0.1'),
                possible_branch_names=('main',),
                tag_msg_log_format='%s',
            )
            self.assertIs(pgit.git_tag_msg, None)
            pgit.fast_checks()
            self.assertEqual(
                pgit.git_tag_msg,
                'publish version v0.0.1 with these changes:\n\n * The initial commit ;)\n',
            )

    def test_publisher_git(self):
        with AssertLogs(self, loggers=('manageprojects',)), TemporaryDirectory(
            prefix='test_publisher_git'
        ) as temp_path:
            Path(temp_path, '1.txt').touch()
            git, first_hash = init_git(temp_path, comment='The initial commit ;)')

            git.tag('v0.0.1', message='one', verbose=False)

            Path(temp_path, '2.txt').touch()
            git.add(spec='.')
            git.commit(comment='Useless 1')

            git.tag('v0.2.0', message='two', verbose=False)

            Path(temp_path, '3.txt').touch()
            git.add(spec='.')
            git.commit(comment='New commit 1')

            Path(temp_path, '4.txt').touch()
            git.add(spec='.')
            git.commit(comment='New commit 2')

            pgit = PublisherGit(
                package_path=temp_path,
                version=Version('0.3.0'),
                possible_branch_names=('main',),
                tag_msg_log_format='%s',
            )
            self.assertIs(pgit.git_tag_msg, None)
            pgit.fast_checks()
            self.assertEqual(
                pgit.git_tag_msg,
                'publish version v0.3.0 with these changes (since v0.2.0):\n\n * New commit 2\n * New commit 1\n',
            )

            def return_callback(popenargs, args, kwargs):
                git_command = popenargs[1]
                if git_command == 'fetch':
                    return
                elif git_command == 'log':
                    return FakeStdout(stdout='')
                elif git_command == 'push':
                    return FakeStdout(stdout='Everything up-to-date')

            with SubprocessCallMock(return_callback) as call_mock:
                pgit.slow_checks()

            self.assertEqual(
                call_mock.get_popenargs(rstrip_paths=(GIT_BIN_PARENT,)),
                [
                    ['.../git', 'fetch', 'origin'],
                    ['.../git', 'log', 'HEAD..origin/main', '--oneline'],
                    ['.../git', 'push', 'origin', 'main'],
                ],
            )

            with SubprocessCallMock() as call_mock:
                pgit.finalize()

            self.assertEqual(
                call_mock.get_popenargs(rstrip_paths=(GIT_BIN_PARENT,)),
                [
                    [
                        '.../git',
                        'tag',
                        '-a',
                        'v0.3.0',
                        '-m',
                        (
                            'publish version v0.3.0 with these changes (since v0.2.0):'
                            '\n\n * New commit 2\n * New commit 1\n'
                        ),
                    ],
                    ['.../git', 'push', '--tags'],
                ],
            )



import logging
import shutil
from pathlib import Path
from typing import Optional

from bx_py_utils.path import assert_is_dir
from rich import print

from manageprojects.cookiecutter_api import execute_cookiecutter
from manageprojects.data_classes import GenerateTemplatePatchResult
from manageprojects.git import Git
from manageprojects.utilities.temp_path import TemporaryDirectory


logger = logging.getLogger(__name__)


def verbose_copy(src, dst):
    logger.info(f'copy: "{src}" to "{dst}"')
    shutil.copy2(src, dst)


def make_git_diff(temp_path: Path, from_path: Path, to_path: Path, verbose=True) -> Optional[str]:
    """
    Create git diff between from_path and to_path
    """
    print(f'Make git diff between {from_path} and {to_path} (temp: {temp_path})')
    assert_is_dir(from_path)
    assert_is_dir(to_path)

    temp_repo_path = temp_path / 'git-repo'

    # Create git repo with "from" content:
    shutil.copytree(
        src=from_path,
        dst=temp_repo_path,
        ignore=shutil.ignore_patterns('.git'),
        copy_function=verbose_copy,
        dirs_exist_ok=False,
    )
    assert not Path(temp_repo_path, '.git').exists()
    git = Git(cwd=temp_repo_path, detect_root=False)
    git.init(verbose=verbose)
    git.add('.', verbose=False)
    git.commit('init with "from" revision', verbose=False)
    git.print_file_list(out_func=logger.info)

    # Remove all files, except .git:
    for item in temp_repo_path.iterdir():
        if item.is_dir():
            if item.name == '.git':
                continue
            shutil.rmtree(item)
        else:
            item.unlink()

    # Commit "to" version:
    shutil.copytree(
        src=to_path,
        dst=temp_repo_path,
        ignore=shutil.ignore_patterns('.git'),
        copy_function=verbose_copy,
        dirs_exist_ok=True,
    )
    git.add('.', verbose=verbose)
    git.commit('Commit "to" revision', verbose=verbose)
    git.print_file_list(out_func=logger.info)

    # Diff between previous commit (from) and current commit (to):
    patch = git.diff('HEAD^', 'HEAD')
    if not patch:
        logger.warning(f'No gif diff between {from_path} and {to_path} !')
        return None

    return patch


def generate_template_patch(
    *,
    project_path: Path,
    template: str,  # CookieCutter Template path or GitHub url
    from_rev: str,
    replay_context: dict,
    directory: Optional[str] = None,  # Directory name of the CookieCutter Template
    password: Optional[str] = None,
    config_file: Optional[Path] = None,  # Optional path to 'cookiecutter_config.yaml'
    cleanup: bool = True,  # Remove temp files if not exceptions happens
    no_input: bool = False,  # Prompt the user at command line for manual configuration?
) -> Optional[GenerateTemplatePatchResult]:
    """
    Create git diff/patch from cookiecutter template changes.
    """
    print(f'Generate update patch for project: {project_path} from {template}')
    project_name = project_path.name

    extra_context = replay_context.get('cookiecutter')
    if not extra_context:
        print('WARNING: No "cookiecutter" in replay context!')

    with TemporaryDirectory(prefix=f'manageprojects_{project_name}_', cleanup=cleanup) as temp_path:

        #############################################################################
        # Generate the cookiecutter template in the current/HEAD version:

        compiled_to_path = temp_path / 'to_rev_compiled'
        print(f'Compile cookiecutter template in the current version here: {compiled_to_path}')
        print('Use extra context:')
        print(extra_context)
        to_rev_context, to_rev_dst_path, to_rev_repo_path = execute_cookiecutter(
            template=template,
            directory=directory,
            output_dir=compiled_to_path,
            no_input=no_input,
            extra_context=extra_context,
            checkout=None,  # Checkout HEAD/main revision
            password=password,
            config_file=config_file,
        )

        assert_is_dir(to_rev_repo_path)
        assert to_rev_dst_path.parent == compiled_to_path

        #############################################################################
        # Get the current git commit hash and date:

        git = Git(cwd=to_rev_repo_path, detect_root=True)
        to_rev = git.get_current_hash(verbose=False)
        to_commit_date = git.get_commit_date(verbose=False)
        print(f'Update from rev. {from_rev} to rev. {to_rev} ({to_commit_date})')

        if from_rev == to_rev:
            print(
                f'Latest version {from_rev!r}'
                f' from {to_commit_date} is already applied.'
                ' Nothing to update, ok.'
            )
            return None

        if 'github.com' in template:
            print(f'Github compare: {template}/compare/{from_rev}...{to_rev}')

        patch_file_path = Path(
            project_path, '.manageprojects', 'patches', f'{from_rev}_{to_rev}.patch'
        )
        print(f'Generate patch file: {patch_file_path}')

        #############################################################################
        # Generate the cookiecutter template in the old version:

        compiled_from_path = temp_path / f'{from_rev}_compiled'
        print(
            'Compile cookiecutter template in the'
            f' old {from_rev} version here: {compiled_from_path}'
        )
        print('Use extra context:')
        print(extra_context)
        from_rev_context, from_rev_dst_path, from_repo_path = execute_cookiecutter(
            template=template,
            directory=directory,
            output_dir=compiled_from_path,
            no_input=no_input,
            extra_context=extra_context,
            checkout=from_rev,  # Checkout the old revision
            password=password,
            config_file=config_file,
        )
        assert_is_dir(from_repo_path)
        assert from_repo_path == to_rev_repo_path
        assert from_rev_dst_path.parent == compiled_from_path

        #############################################################################
        # Generate git patch between old and current version:

        patch = make_git_diff(
            temp_path=temp_path,
            from_path=from_rev_dst_path,
            to_path=to_rev_dst_path,
            verbose=False,
        )
        if not patch:
            print(f'No gif diff between {compiled_from_path} and {compiled_to_path} !')
            return None

        logger.info('Write patch file: %s', patch_file_path)
        patch_file_path.parent.mkdir(parents=True, exist_ok=True)
        patch_file_path.write_text(patch)

        return GenerateTemplatePatchResult(
            repo_path=to_rev_repo_path,  # == from_repo_path
            patch_file_path=patch_file_path,
            from_rev=from_rev,
            compiled_from_path=compiled_from_path,
            to_rev=to_rev,
            to_commit_date=to_commit_date,
            compiled_to_path=compiled_to_path,
        )



import dataclasses
import datetime
import logging
from pathlib import Path
from typing import Optional

import tomlkit
from bx_py_utils.path import assert_is_dir, assert_is_file
from tomlkit import TOMLDocument
from tomlkit.items import Table

from manageprojects.constants import (
    APPLIED_MIGRATIONS,
    COOKIECUTTER_CONTEXT,
    COOKIECUTTER_DIRECTORY,
    COOKIECUTTER_TEMPLATE,
    INITIAL_DATE,
    INITIAL_REVISION,
)
from manageprojects.data_classes import ManageProjectsMeta
from manageprojects.exceptions import NoPyProjectTomlFound
from manageprojects.utilities.log_utils import log_func_call


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class TomlDocument:
    file_path: Path
    doc: TOMLDocument


def get_toml_document(path: Path) -> TomlDocument:
    """
    Read the file path and returns the TOMLDocument.
    Just use unwrap() if a normal dict is needed, e.g.:

        doc: TOMLDocument = get_toml_document(path)
        toml: dict = doc.unwrap()
    """
    assert_is_file(path)
    doc: TOMLDocument = tomlkit.parse(path.read_text(encoding='UTF-8'))

    return TomlDocument(file_path=path, doc=doc)


def find_pyproject_toml(file_path: Path) -> Optional[Path]:
    """
    Go back down the directory tree to find the "pyproject.toml" file.
    """
    if file_path.is_file():
        return find_pyproject_toml(file_path.parent)

    pyproject_toml = file_path / 'pyproject.toml'
    if pyproject_toml.is_file():
        return pyproject_toml

    if parent_path := file_path.parent:
        if parent_path != file_path:
            return find_pyproject_toml(parent_path)

    return None


def get_pyproject_toml(*, file_path: Optional[Path] = None) -> TomlDocument:
    """
    Find the "pyproject.toml" and return it.
    """
    if not file_path:
        file_path = Path.cwd()
    pyproject_toml_path = find_pyproject_toml(file_path=file_path)
    if not pyproject_toml_path:
        raise NoPyProjectTomlFound(f'Can not find "pyproject.toml" in {file_path}')

    assert_is_file(pyproject_toml_path)
    toml_document: TomlDocument = get_toml_document(pyproject_toml_path)
    return toml_document


class PyProjectToml:
    """
    Helper for manageprojects meta information in 'pyproject.toml'
    """

    def __init__(self, project_path: Path):
        assert_is_dir(project_path)
        self.path = project_path / 'pyproject.toml'

        if self.path.exists():
            logger.debug('Read existing pyproject.toml')
            self.doc: TOMLDocument = tomlkit.parse(self.path.read_text(encoding='UTF-8'))
        else:
            logger.debug('Create new pyproject.toml')
            self.doc: TOMLDocument = tomlkit.document()  # type: ignore
            self.doc.add(tomlkit.comment('Created by manageprojects'))

        self.mp_table: Table = self.doc.get('manageprojects')  # type: ignore
        if not self.mp_table:
            # Insert: [manageprojects]
            if self.path.exists():
                self.doc.add(tomlkit.ws('\n\n'))  # Add a new empty line
            self.mp_table = tomlkit.table()
            self.mp_table.comment('https://github.com/jedie/manageprojects')
            self.doc.append('manageprojects', self.mp_table)

    def init(
        self, revision, dt: datetime.datetime, template: str, directory: Optional[str]
    ) -> None:
        assert INITIAL_REVISION not in self.mp_table
        assert INITIAL_DATE not in self.mp_table
        assert COOKIECUTTER_TEMPLATE not in self.mp_table
        self.mp_table.add(INITIAL_REVISION, revision)
        self.mp_table.add(INITIAL_DATE, dt)
        self.mp_table.add(COOKIECUTTER_TEMPLATE, template)
        if directory:
            self.mp_table.add(COOKIECUTTER_DIRECTORY, directory)

    def create_or_update_cookiecutter_context(self, context: dict) -> None:
        # Don't store the '_output_dir':
        context['cookiecutter'] = {
            k: v for k, v in context['cookiecutter'].items() if k != '_output_dir'
        }
        self.mp_table[COOKIECUTTER_CONTEXT] = tomlkit.item(context)

    def add_applied_migrations(self, git_hash: str, dt: datetime.datetime) -> None:
        if not (applied_migrations := self.mp_table.get(APPLIED_MIGRATIONS)):
            # Add: applied_migrations = []
            applied_migrations = tomlkit.array()
            applied_migrations.multiline(multiline=True)
            self.mp_table.add(APPLIED_MIGRATIONS, applied_migrations)

        # Append git_hash to applied_migrations:
        applied_migrations.add_line(git_hash, comment=dt.isoformat())

    ###############################################################################################

    def dumps(self) -> str:
        return self.doc.as_string()

    def save(self) -> None:
        content = self.dumps()
        self.path.write_text(content, encoding='UTF-8')

    ###############################################################################################

    def get_mp_meta(self) -> ManageProjectsMeta:
        data = self.mp_table.unwrap()  # change tomlkit.Container to a normal dict
        result: ManageProjectsMeta = log_func_call(
            logger=logger,
            func=ManageProjectsMeta,
            initial_revision=data.get(INITIAL_REVISION),
            initial_date=data.get(INITIAL_DATE),
            applied_migrations=data.get(APPLIED_MIGRATIONS, []),
            cookiecutter_template=data.get(COOKIECUTTER_TEMPLATE),
            cookiecutter_directory=data.get(COOKIECUTTER_DIRECTORY),
            cookiecutter_context=data.get(COOKIECUTTER_CONTEXT),
        )
        return result



import tempfile
from pathlib import Path
from unittest import mock
from unittest.mock import MagicMock


from manageprojects.cli import cli_app
from manageprojects.cli.cli_app import PACKAGE_ROOT, cli, start_project, update_project
from manageprojects.constants import PY_BIN_PATH
from manageprojects.data_classes import CookiecutterResult
from manageprojects.test_utils.click_cli_utils import invoke_click
from manageprojects.test_utils.subprocess import SubprocessCallMock
from manageprojects.tests.base import BaseTestCase


class CliTestCase(BaseTestCase):

    def test_start_project_cli(self):
        result = CookiecutterResult(
            destination_path=Path('destination'),
            git_path=Path('git-path'),
            git_hash='1234',
            commit_date=None,
            cookiecutter_context=dict(foo=1, bar=2),
        )
        with mock.patch.object(cli_app, 'start_managed_project', MagicMock(return_value=result)) as m:
            stdout = invoke_click(
                start_project,
                'https://github.com/jedie/cookiecutter_templates/',
                '--directory',
                'piptools-python',
                'foobar/',
            )
        m.assert_called_once_with(
            template='https://github.com/jedie/cookiecutter_templates/',
            directory='piptools-python',
            output_dir=Path('foobar'),
            checkout=None,
            input=True,
            replay=False,
            password=None,
            config_file=None,
        )
        self.assert_in_content(
            got=stdout,
            parts=(
                'https://github.com/jedie/cookiecutter_templates/',
                'foobar',
                'git hash 1234',
            ),
        )

    def test_update_project_cli(self):
        tempdir = tempfile.gettempdir()

        with mock.patch.object(cli_app, 'update_managed_project') as m:
            stdout = invoke_click(update_project, tempdir)

        m.assert_called_once_with(
            project_path=Path(tempdir),
            overwrite=True,
            password=None,
            config_file=None,
            cleanup=True,
            input=False,
        )
        self.assert_in_content(
            got=stdout,
            parts=(
                f'Update project: "{tempdir}"...',
                f'Managed project "{tempdir}" updated, ok.',
            ),
        )

    def test_install(self):
        with SubprocessCallMock() as call_mock:
            invoke_click(cli, 'install')

        self.assertEqual(
            call_mock.get_popenargs(rstrip_paths=(PY_BIN_PATH,)),
            [
                ['.../pip-sync', f'{PACKAGE_ROOT}/requirements.dev.txt'],
                ['.../pip', 'install', '--no-deps', '-e', '.'],
            ],
        )
